"""This module produces maps for a single observation of a bright
point source.  The observation is identifed by an obs_id.  The data
for the observation may be divided into different detector groups; and
each 'group' will be loaded and mapped independently (this will
normally be associated with a "detset").  The data for each
observation in each group may be further subdivided into 'data
splits'; this normally corresponds to frequency "band".

"""

from argparse import ArgumentParser
import logging
import numpy as np
import os
import sys
import yaml

import sotodlib
import so3g
from sotodlib import core, coords, site_pipeline, tod_ops

from . import util

logger = None


def get_parser(parser=None):
    if parser is None:
        parser = ArgumentParser()
    parser.add_argument('-c', '--config-file', help=
                        "Configuration file.")
    parser.add_argument('-v', '--verbose', action='count',
                        default=0, help="Pass multiple times to increase.")
    parser.add_argument('obs_id',help=
                        "Observation for which to make source map.")
    parser.add_argument('--test', action='store_true', help=
                        "Reduce detector count for quick tests.")

    return parser

def _get_config(config_file):
    return yaml.safe_load(open(config_file, 'r'))

def _clip_map(m, mask=None, edges=[0.05, 0.95]):
    if mask is None:
        mask = ~np.isnan(m)
    lims = np.quantile(m[mask], edges)
    lims = lims[0] + (lims[1] - lims[0]) * \
           (np.array([0., 1.]) - edges[0]) / (edges[1] - edges[0])
    out = np.clip(m, *lims)
    out[~mask] = np.nan
    return out

def _renorm(im):
    s = ~np.isnan(im)
    mag = abs(im[s]).max()
    if mag <= 0:
        alpha = 0
    else:
        alpha = int(np.floor(np.log10(mag)))
    rescale = 10**-alpha
    label = '$10^{%d}$' % alpha
    imr = im.copy()
    imr[s] *= rescale
    return imr, label, rescale

def _squarify(ax, lims=None):
    if lims is None:
        lims = []
        xl = ax.get_xlim()
        yl = ax.get_ylim()
        for z in [xl, yl]:
            z0, dz = (z[0] + z[1])/2, (z[0] - z[1])/2
            lims.append([z0, np.sign(dz), abs(dz)])
        dz = max([l[2] for l in lims])
        lims = [(l[0] + l[1]*dz, l[0] - l[1]*dz) for l in lims]
    ax.set_xlim(*lims[0])
    ax.set_ylim(*lims[1])

def _plot_one_map(fig, ax, m):
    mr, units, _ = _renorm(m)
    im = ax.imshow(mr)
    fig.colorbar(im, ax=ax, label=f'Units ({units})',
                 orientation='horizontal')
    return im

def plot_map(bundle, filename=None, tod=None, obs_info=None, det_info=None,
             focal_plane=None, det_mask=None, group=None, subset=None,
             zoom_size=None, title=None, **kwargs):
    import matplotlib.pyplot as plt

    if tod is not None:
        if obs_info is None:
            obs_info = tod.get('obs_info')
        if det_info is None:
            det_info = tod.get('det_info')
        if focal_plane is None:
            focal_plane = tod.get('focal_plane')

    src_map = bundle['solved']
    if src_map.shape[0] == 3:
        fig, axs = plt.subplots(2, 3, figsize=(9, 8),
                                subplot_kw={'projection': 'so-beammap'})
        axs = {
            'W':  axs[0, 0],
            'T':  axs[0, 1],
            'fp': axs[0, 2],
            'Tz': axs[1, 0],
            'Qz': axs[1, 1],
            'Uz': axs[1, 2],
        }

    else:
        fig, axs = plt.subplots(2, 2, figsize=(7, 8),
                                subplot_kw={'projection': 'so-beammap'})
        axs = {
            'W':  axs[0, 0],
            'T':  axs[0, 1],
            'Tz': axs[1, 0],
            'fp': axs[1, 1],
        }

    mask = bundle['weights'][0,0] != 0
    _plot_one_map(fig, axs['W'],
                  _clip_map(bundle['weights'][0, 0], mask))
    _plot_one_map(fig, axs['T'],
                  _clip_map(src_map[0], mask))
    _squarify(axs['W'])
    _squarify(axs['T'])

    # zoomed in ...
    if zoom_size is None:
        zoom_size = 0.5  # deg
    Z = zoom_size * coords.DEG
    box = np.array([[-Z, Z], [Z, -Z]])
    zoomed = src_map.submap(box)

    # Plot whichever of T/Q/U are appearing here.
    for i, k in enumerate(['Tz', 'Qz', 'Uz']):
        if k in axs:
            _plot_one_map(fig, axs[k], zoomed[i])

    if focal_plane:
        x, y = focal_plane.xi / coords.DEG, focal_plane.eta / coords.DEG
        if det_mask is not None:
            x, y = x[det_mask], y[det_mask]
        axs['fp'].scatter(x, y, s=2, alpha=.25)
        axs['fp'].set_aspect('equal')

    if title is None:
        try:
            title = '{obs_info.obs_id} - {group} - {subset}'.format(
                obs_info=obs_info, group=group, subset=subset)
        except:
            title = ''
    plt.suptitle(title)

    for k, label, c in [
            ('W', 'Weight', 'k'),
            ('T', 'Signal', 'k'),
            ('Tz', 'T', 'w'),
            ('Qz', 'Q', 'w'),
            ('Uz', 'U', 'w')]:
        if k in axs:
            axs[k].text(0.95, 0.95, label, color=c,
                        horizontalalignment='right',
                        verticalalignment='top',
                        transform=axs[k].transAxes)

    plt.tight_layout()
    plt.subplots_adjust(top=.9)

    if filename is not None:
        fig.savefig(filename)
    return fig

def _adjust_focal_plane(tod, focal_plane=None, boresight_offset=None):
    """Apply pointing correction to focal plane.  These are assumed to be
    boresight corrections, even though there is one per detector.

    """
    if focal_plane is None:
        focal_plane = tod.focal_plane
    if boresight_offset is None:
        boresight_offset = tod.boresight_offset

    # Get detector the boresight pointing quaternions
    fp = focal_plane
    fp_q = so3g.proj.quat.rotation_xieta(fp.xi, fp.eta, fp.gamma)

    # Get the adjustments
    bc = boresight_offset
    fp_adjust = so3g.proj.quat.rotation_xieta(bc.dx, bc.dy, bc.gamma)

    # Apply focal plane adjust.
    fp_new = fp_adjust * fp_q

    # Get modified xi, eta, gamma.
    xi, eta, gamma = so3g.proj.quat.decompose_xieta(fp_new)
    fp.xi[:] = xi
    fp.eta[:] = eta
    fp.gamma[:] = gamma

def main(config_file=None, obs_id=None, verbose=0, test=False):
    """Entry point."""
    config = _get_config(config_file)

    logger = util.init_logger(__name__, 'make_uncal_beam_map: ')
    if verbose >= 1:
        logger.setLevel('INFO')
    if verbose >= 2:
        sotodlib.logger.setLevel('INFO')
    if verbose >= 3:
        sotodlib.logger.setLevel('DEBUG')

    ctx = core.Context(config['context_file'])

    group_by = config['subobs'].get('use', 'detset')
    if group_by.startswith('dets:'):
        group_by = group_by.split(':',1)[1]

    if group_by == 'detset':
        groups = ctx.obsfiledb.get_detsets(obs_id)
    else:
        det_info = ctx.get_det_info(obs_id)
        groups = det_info.subset(keys=[group_by]).distinct()[group_by]

    # Ignore some values?
    groups = [g for g in groups if g not in config['subobs'].get('ignore_vals', [])]

    if len(groups) == 0:
        logger.warning(f'No map groups found for obs_id={obs_id}')

    if os.path.exists(config['archive']['index']):
        logger.info(f'Mapping {config["archive"]["index"]} for the archive index.')
        db = core.metadata.ManifestDb(config['archive']['index'])
    else:
        logger.info(f'Creating {config["archive"]["index"]} for the archive index.')
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('group')
        scheme.add_data_field('split')
        db = core.metadata.ManifestDb(config['archive']['index'], scheme=scheme)

    for group in groups:
        logger.info(f'Loading {obs_id}:{group_by}={group}')
        map_info = {'obs_id': obs_id,
                    'group': group,
                    group_by: group,
        }
        map_info['product_id'] = '{obs_id}-{group}'.format(**map_info)

        # Read data.
        tod = ctx.get_obs(obs_id, dets={group_by: group})
        if len(tod.signal) == 0:
            logger.warning(f'No detectors loaded for obs_id={obs_id}, '
                           f'group {group_by}={group}; skipping.')
            continue

        # Modify dets axis for testing
        if test:
            logger.warning(f'Decimating focal plane (--test).')
            dets = tod.dets.vals[::10]
            tod.restrict('dets', dets)

        # Modify samps axis for FFTs.
        logger.info(f' -- Before trimming, TOD shape is: {tod.shape}.')
        tod_ops.fft_trim(tod)
        logger.info(f' -- After trimming, TOD shape is:  {tod.shape}.')

        # Determine what source to map.
        ## To-do: have a mode where it maps all sources it finds.
        source_name = config['mapmaking'].get('force_source')
        if source_name is None:
            sources = coords.planets.get_nearby_sources(tod)
            if len(sources) == 1:
                source_name = sources[0][0]
            elif len(sources) == 0:
                logger.error("No mappable sources in footprint.")
                sys.exit(1)
            else:
                logger.error("Multiple sources in footprint: %s" %
                             ([n for n, s in sources],))
                sys.exit(1)
        map_info['source_name'] = source_name

        # Plan to split on frequency band
        band_splits = coords.planets.load_detector_splits(tod, source=tod.det_info['band'])

        # Deconvolve readout filter and detector time constants.
        tod_ops.fft_trim(tod)
        tod_ops.detrend_tod(tod)

        filt = tod_ops.filters.identity_filter()
        if 'iir_params' in tod:
            filt *= tod_ops.filters.iir_filter(invert=True)
        if 'timeconst' in tod:
            filt *= tod_ops.filters.timeconst_filter(invert=True)

        tod.signal[:] = tod_ops.fourier_filter(tod, filt, resize=None, detrend=None)

        # Demodulation & downsampling.
        # ...

        # Fix pointing.
        logger.info('Applying pointing corrections.')
        if 'boresight_offset' in tod:
            logger.info(' -- applying "boresight_offset".')
            _adjust_focal_plane(tod)

        # Apply calibration
        cal = None
        for k in config['preprocessing']['cal_keys']:
            if k in tod:
                if cal is None:
                    cal = 1.
                cal *= tod[k]
        if cal is not None:
            tod.signal *= cal[:,None]

        # Figure out the resolution
        reses = [
            util.lookup_conditional(config['mapmaking'], 'res', tags=[b])
            for b in band_splits.keys()]
        assert(all([r == reses[0] for r in reses]))  # Resolution conflict
        res_deg = util.parse_quantity(reses[0], 'deg').value

        # Where to put things.
        policy = util.ArchivePolicy.from_params(
            config['archive']['policy'])

        dest_dir = policy.get_dest(**map_info)
        if os.path.exists(dest_dir):
            logger.info(f' -- destination already exists ({dest_dir})')
        else:
            logger.info(f' -- creating destination dir ({dest_dir})')
            os.makedirs(dest_dir)

        # Make the maps.
        logger.info(f'Calling mapmaker.')
        cuts = tod.flags.get('glitch_flags', None)
        output = coords.planets.make_map(tod, center_on=source_name,
                                         res=res_deg*coords.DEG,
                                         data_splits=band_splits,
                                         cuts=cuts,
                                         info=map_info,
                                         thread_algo='domdir')

        # Save and plot
        ocfg = config['output']
        pcfg = config['plotting']
        opattern = ocfg['pattern']
        ppattern = pcfg.get('pattern', opattern + '.png')

        used_names = []  # watch for accidental self-overwrites
        db_code = ocfg['map_codes'][0]

        for key, bundle in output['splits'].items():
            logger.info(f'Processing outputs for group={group}, band={key}...')
            # Save
            for code in ocfg['map_codes']:
                filename = os.path.join(
                    dest_dir, opattern.format(map_code=code, split=key, **map_info))
                logger.info(f' -- writing map to {filename}')
                bundle[code].write(filename)
                if filename in used_names:
                    logger.warning(f'Wrote this file more than once: {filename}; '
                                   f'using {key}, {code}')
                used_names.append(filename)
                if code == db_code:
                    # Index
                    db_data = {'obs:obs_id': obs_id,
                               'group': group,
                               'split': key}
                    db.add_entry(db_data, filename, replace=True)

            # Plot
            plot_filename = os.path.join(
                dest_dir, ppattern.format(map_code='solved', split=key, **map_info))
            logger.info(f' -- writing plots to {plot_filename}...')

            band_mask = (tod.det_info['band'] == key)
            zoom_size = util.parse_quantity(
                util.lookup_conditional(pcfg, 'zoom', tags=[key]),
                'deg').value

            plot_map(bundle, plot_filename, zoom_size=zoom_size,
                     tod=tod, det_mask=band_mask, group=group, subset=key)


    # Return something?
    return True


if __name__ == '__main__':
    util.main_launcher(main, get_parser)



"""
Command line / prefect tool for working with an hkdb.
"""


import argparse

from sotodlib.io import hkdb
from sotodlib.site_pipeline import util

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument("cfg_file", help="Path to update_hkdb.yaml config file.")
    parser.add_argument('--verbose', action='store_true')

    sps = parser.add_subparsers(title="command", dest="command")

    # list
    sp = sps.add_parser('list', help="Summarize files known to the database.")
    sp.add_argument('--all', action='store_true', help=
                    "Show all records (otherwise limited to 10).")
    sp.add_argument('--compact', action='store_true', help=
                    "Show all records in compact format.")
    sp.add_argument('--status', action='append', help=
                    "Limit what types of records are shown.",
                    choices=['unindexed', 'indexed', 'failed'])

    # update
    sp = sps.add_parser('update', help="Scan for new files, and run indexing.")
    sp.add_argument('--lookback-days', type=float, help=
                    "Set the lookback time window in days; overrides any "
                    "file_idx_lookback_time setting in config file.")
    sp.add_argument('--retry-failed', action='store_true', default=False,
                    help="Also retry any files that are known, but 'failed'.")

    # reset-files
    sp = sps.add_parser('reset-files', help="Mark all 'failed' files as "
                        "unindexed, which will cause them to be rescanned. "
                        "Use --pattern to restrict to a subset.")
    sp.add_argument('--pattern', '-p', help=
                    "sql pattern to match with (e.g. '%_17244_%')")

    return parser


def cli_main(**kw):
    cmd = kw.get('command')
    cfg = hkdb.HkConfig.from_yaml(kw['cfg_file'])

    if kw.get('verbose'):
        util.init_logger('', logger=hkdb.log, verbosity=3)

    if cmd is None:
        print('Pass -h to get usage.')

    elif cmd == 'list':
        limit = 10
        if kw.get('all') or kw.get('compact'):
            limit = None
        index_status = kw.get('status')
        report = hkdb.get_files_info(cfg, limit=limit, index_status=index_status)
        if kw.get('compact'):
            last_r = None
            same = 0
            fmt = '  {r.path} {r.index_status}'
            for r in report:
                if last_r is None or r.index_status != last_r.index_status:
                    if same > 1:
                        print(' ... and %i more %s' % (same - 1, last_r.index_status))
                        print(fmt.format(r=last_r))
                    print(fmt.format(r=r))
                    same = 0
                else:
                    same += 1
                last_r = r
            if same > 1:
                print(' ... and %i more %s' % (same - 1, last_r.index_status))
                print(fmt.format(r=r))
        else:
            for r in report:
                print(f'{r.path} {r.index_status}')
            if limit is not None and len(report) >=  limit:
                print(' .... and possibly others (try --all or --compact)')

    elif cmd == 'update':
        if kw.get('lookback_days'):
            cfg.file_idx_lookback_time = kw['lookback_days'] * 86400
        report = hkdb.update_index_all(cfg, retry_failed=kw.get('retry_failed'))
        print('Report')
        for k, v in report.items():
            print(f'  {k}: {v}')

    elif cmd == 'reset-files':
        n = hkdb.reset_failed_files(cfg, pattern=kw.get('pattern'))
        print(f'Changed {n} entries.')


def simple_update(cfg_file: str, retry_failed: bool=False):
    """Updates hkdb index databases"""
    util.init_logger('', logger=hkdb.log, verbosity=3)
    stats = hkdb.update_index_all(cfg_file, retry_failed=retry_failed)
    hkdb.log.info(f'hkdb update summary: {stats}')
    if stats['new_files_failed']:
        raise RuntimeError('Some indexing failures encountered.')


# For prefect, map main to simple_update.
main = simple_update


if __name__ == "__main__":
    parser = get_parser(parser=None)
    args = parser.parse_args()
    cli_main(**vars(args))



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulated detector response to sky signal.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from ..instrument import simulated_telescope
from .job import workflow_timer


def setup_simulate_sky_map_signal(operators):
    """Add commandline args and operators for scanning from a map.

    There are pointing matrix operators associated with each scan map operator
    and these are disabled by default.  If not enabled, the scan map operator
    use the same pointing matrix as the mapmaking.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.ScanHealpixMap(name="scan_map", enabled=False))
    operators.append(
        toast.ops.StokesWeights(
            name="scan_map_weights",
            mode="IQU",
            weights="weights_scan_map",
            enabled=False,
        )
    )
    operators.append(
        toast.ops.PixelsHealpix(
            name="scan_map_pixels", pixels="pixels_scan_map", enabled=False
        )
    )
    operators.append(toast.ops.ScanWCSMap(name="scan_wcs_map", enabled=False))
    operators.append(
        toast.ops.StokesWeights(
            name="scan_wcs_map_weights",
            mode="IQU",
            weights="weights_scan_map",
            enabled=False,
        )
    )
    operators.append(
        toast.ops.PixelsWCS(
            name="scan_wcs_map_pixels",
            projection="CAR",
            pixels="pixels_scan_map",
            resolution=(0.005 * u.degree, 0.005 * u.degree),
            submaps=1,
            auto_bounds=True,
            enabled=False,
        )
    )


@workflow_timer
def simulate_sky_map_signal(job, otherargs, runargs, data):
    """Scan a sky map into detector signals.

    This uses detector pointing to sample from a map into detector
    timestreams.  The maps should already be smoothed with any desired
    beam effects.

    We scan the sky with the "final" pointing model in case that is
    different from the solver pointing model.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    # Configured operators for this job
    job_ops = job.operators

    if job_ops.scan_map.enabled and job_ops.scan_wcs_map.enabled:
        msg = "Cannot scan from both healpix and WCS maps"
        log.error(msg)
        raise RuntimeError(msg)

    if job_ops.det_pointing_radec_sim is not job_ops.det_pointing_radec:
        if job_ops.scan_map.enabled and not job_ops.scan_map_pixels.enabled:
            msg = "Simulation pointing is different from data reduction pointing. " \
                " You must enable scan_map_pixels"
            log.error(msg)
            raise RuntimeError(msg)
        if job_ops.scan_wcs_map.enabled and not job_ops.scan_wcs_map_pixels.enabled:
            msg = "Simulation pointing is different from data reduction pointing. " \
                " You must enable scan_wcs_map_pixels"
            log.error(msg)
            raise RuntimeError(msg)

    if job_ops.scan_map.enabled:
        if job_ops.scan_map_pixels.enabled:
            # We are using a custom pointing matrix
            job_ops.scan_map_pixels.detector_pointing = job_ops.det_pointing_radec_sim
            job_ops.scan_map.pixel_dist = "scan_map_pixel_dist"
            job_ops.scan_map.pixel_pointing = job_ops.scan_map_pixels
        else:
            # We are using the same pointing matrix as the mapmaking
            job_ops.scan_map.pixel_dist = job_ops.binner_final.pixel_dist
            job_ops.scan_map.pixel_pointing = job.pixels_final
        if job_ops.scan_map_weights.enabled:
            job_ops.scan_map_weights.detector_pointing = job_ops.det_pointing_radec_sim
            job_ops.scan_map.stokes_weights = job_ops.scan_map_weights
        else:
            job_ops.scan_map.stokes_weights = job_ops.weights_radec
        job_ops.scan_map.save_pointing = otherargs.full_pointing
        job_ops.scan_map.apply(data)
        if job_ops.scan_map_pixels.enabled:
            # Clean up our custom pointing
            toast.ops.Delete(detdata=[
                job_ops.scan_map_pixels.pixels,
                job_ops.det_pointing_radec_sim.quats,
            ]).apply(data)
        if job_ops.scan_map_weights.enabled:
            # Clean up our custom pointing
            toast.ops.Delete(detdata=[
                job_ops.scan_map_weights.weights,
            ]).apply(data)

    if job_ops.scan_wcs_map.enabled:
        if job_ops.scan_wcs_map_pixels.enabled:
            # We are using a custom pointing matrix
            job_ops.scan_wcs_map_pixels.detector_pointing = job_ops.det_pointing_radec_sim
            job_ops.scan_wcs_map.pixel_dist = "scan_wcs_map_pixel_dist"
            job_ops.scan_wcs_map.pixel_pointing = job_ops.scan_wcs_map_pixels
        else:
            # We are using the same pointing matrix as the mapmaking
            job_ops.scan_wcs_map.pixel_dist = job_ops.binner_final.pixel_dist
            job_ops.scan_wcs_map.pixel_pointing = job.pixels_final
        if job_ops.scan_wcs_map_weights.enabled:
            job_ops.scan_wcs_map_weights.detector_pointing = job_ops.det_pointing_radec_sim
            job_ops.scan_wcs_map.stokes_weights = job_ops.scan_wcs_map_weights
        else:
            job_ops.scan_wcs_map.stokes_weights = job_ops.weights_radec
        job_ops.scan_wcs_map.save_pointing = otherargs.full_pointing
        job_ops.scan_wcs_map.apply(data)
        if job_ops.scan_wcs_map_pixels.enabled:
            # Clean up our custom pointing
            toast.ops.Delete(detdata=[
                job_ops.scan_wcs_map_pixels.pixels,
                job_ops.det_pointing_radec_sim.quats,
            ]).apply(data)
        if job_ops.scan_wcs_map_weights.enabled:
            # Clean up our custom pointing
            toast.ops.Delete(detdata=[
                job_ops.scan_wcs_map_weights.weights,
            ]).apply(data)


def setup_simulate_conviqt_signal(operators):
    """Add commandline args and operators for beam covolution with conviqt.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.SimConviqt(name="conviqt", enabled=False))
    operators.append(toast.ops.SimTEBConviqt(name="conviqt_teb", enabled=False))


@workflow_timer
def simulate_conviqt_signal(job, otherargs, runargs, data):
    """Use libconviqt to generate beam-convolved sky signal timestreams.

    This uses detector pointing and input beam and sky a_lm expansions
    to generate detector timestreams.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    # Configured operators for this job
    job_ops = job.operators

    if job_ops.conviqt.enabled and job_ops.sim_ground.hwp_angle is not None:
        msg = "Data has a half-wave plate.  Use conviqt_teb operator "
        msg += "instead of conviqt."
        log.error(msg)
        raise RuntimeError(msg)

    if job_ops.conviqt_teb.enabled and job_ops.sim_ground.hwp_angle is None:
        msg = "Data has no half-wave plate.  Use conviqt operator "
        msg += "instead of conviqt_teb."
        log.error(msg)
        raise RuntimeError(msg)

    if job_ops.sim_ground.hwp_angle is None:
        if job_ops.conviqt.enabled:
            job_ops.conviqt.comm = data.comm.comm_world
            job_ops.conviqt.detector_pointing = job_ops.det_pointing_radec_sim
            job_ops.conviqt.apply(data)
    else:
        if job_ops.conviqt_teb.enabled:
            job_ops.conviqt_teb.comm = data.comm.comm_world
            job_ops.conviqt_teb.detector_pointing = job_ops.det_pointing_radec_sim
            job_ops.conviqt_teb.hwp_angle = job_ops.sim_ground.hwp_angle
            job_ops.conviqt_teb.apply(data)



#!/usr/bin/env python3

import os
import sys
import unittest

from setuptools import find_packages, setup, Command
import versioneer

# The setup options
setup_opts = dict()

setup_opts["version"] = versioneer.get_version()
setup_opts["packages"] = find_packages(where=".", exclude="tests")

setup_opts["package_data"] = {
    "sotodlib": [
        "toast/ops/data/*"
    ]
}
setup_opts["include_package_data"] = True


# Command Class dictionary.
# Begin with the versioneer command class dictionary.
cmdcls = versioneer.get_cmdclass()

# Class to run unit tests

class SOTestCommand(Command):

    def __init__(self, *args, **kwargs):
        super(SOTestCommand, self).__init__(*args, **kwargs)

    def initialize_options(self):
        Command.initialize_options(self)

    def finalize_options(self):
        Command.finalize_options(self)
        self.test_suite = True

    def mpi_world(self):
        # We could avoid putting this here by defining a custom TextTestRunner
        # that checked the per-process value "under-the-hood" and returned a
        # False value from wasSuccessful() if any process failed.
        comm = None
        if "MPI_DISABLE" not in os.environ:
            try:
                from mpi4py import MPI

                comm = MPI.COMM_WORLD
            except Exception:
                pass
        return comm

    def run(self):
        loader = unittest.TestLoader()
        runner = unittest.TextTestRunner(verbosity=2)
        suite = loader.discover("tests", pattern="test_*.py",
                                top_level_dir=".")
        ret = 0
        local_ret = runner.run(suite)
        if not local_ret.wasSuccessful():
            ret = 1
        comm = self.mpi_world()
        if comm is not None:
            ret = comm.allreduce(ret)
        sys.exit(ret)

# Add our custom test runner
cmdcls["test"] = SOTestCommand

# Add command class to setup options
setup_opts["cmdclass"] = cmdcls

# Do the setup.
setup(**setup_opts)



# Copyright (c) 2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Operator for constructing and subtracting dark templates from timestream data.

This operator assembles a dark template signal from the data and projects it out,
enabling improved analysis of detector timestreams by removing correlated noise.
"""

import os
import pickle

from astropy import units as u
import numpy as np
import scipy.signal
import traitlets

import toast
from toast.traits import trait_docs, Unicode, Int, Instance, Bool, Float, List
from toast.ops import Operator
from toast.utils import Logger, Environment, rate_from_times
from toast.timing import function_timer, Timer
from toast.observation import default_values as defaults


@trait_docs
class DarkTemplate(Operator):
    """Operator which assembles and projects out the dark template signal

    """

    API = Int(0, help="Internal interface version for this operator")

    naverage = Int(
        1000,
        help="Lowpass kernel size",
    )

    det_data = Unicode(
        defaults.det_data, help="Observation detdata key for the timestream data"
    )

    det_mask = Int(
        defaults.det_mask_invalid,
        help="Bit mask value for per-detector flagging",
    )

    det_flags = Unicode(
        None, allow_none=True, help="Observation detdata key for flags to use"
    )

    det_flag_mask = Int(
        defaults.det_mask_invalid, help="Bit mask value for optional detector flagging"
    )

    shared_flags = Unicode(
        None, allow_none=True, help="Observation shared key for telescope flags to use"
    )

    shared_flag_mask = Int(
        defaults.shared_mask_invalid, help="Bit mask value for optional shared flagging"
    )

    view = Unicode(
        None, allow_none=True, help="Use this view of the data in all observations"
    )

    key = Unicode(
        "dark_templates",
        help="Shared data key for storing dark bolometer templates",
    )

    cache_dir = Unicode(
        None,
        allow_none=True,
        help="If set, dark templates will be written out or loaded here.",
    )

    @traitlets.validate("det_mask")
    def _check_det_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Det mask should be a positive integer")
        return check

    @traitlets.validate("shared_flag_mask")
    def _check_shared_flag_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Shared flag mask should be a positive integer")
        return check

    @traitlets.validate("det_flag_mask")
    def _check_det_flag_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Det flag mask should be a positive integer")
        return check

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _gather_dark_tod(self, ob):
        """ Gather dark tod to root process

        """
        comm = ob.comm.comm_group
        dark_tod = []
        fp = ob.telescope.focalplane
        for det in ob.select_local_detectors(flagmask=self.det_mask):
            if fp[det]["det_info:wafer:type"] != "DARK":
                continue
            tod = ob.detdata[self.det_data][det]
            # Check for typical pathologies
            if np.any(np.isnan(tod)) or np.std(tod) == 0:
                continue
            # Subtract the mean
            dark_tod.append(tod - np.mean(tod))
        if comm is not None:
            dark_tod = comm.gather(dark_tod)
        return dark_tod

    @function_timer
    def _derive_templates(self, dark_tod):
        """ Derive dark templates from the dark TOD

        """
        dark_tod = np.vstack([x for x in dark_tod if len(x) > 0])
        nsample = dark_tod[0].size
        # PCA
        U, S, Vh = np.linalg.svd(dark_tod, full_matrices=False)
        templates = np.vstack(Vh)
        lowpassed = [np.ones(nsample)]  # always include the offset
        nsum = self.naverage
        lowpass = np.ones(nsum) / nsum
        for template in templates:
            # lowpass each template to suppress uncorrelated high frequency
            # noise.  Start by zero-padding the template to avoid
            # periodicity issues
            embedded = np.zeros(2 * nsample)
            embedded[:nsample] = template
            lowpassed_template = scipy.signal.convolve(
                embedded, lowpass, mode="same"
            )
            lowpassed_template = lowpassed_template[:nsample]
            # Correct the ends for the zero padding (average includes zeros)
            frac = np.ones(template.size)
            frac[:nsum // 2] = nsum / np.arange(nsum // 2, nsum)
            frac[-nsum // 2:] = nsum / np.arange(nsum, nsum // 2, -1)
            lowpassed_template *= frac
            # Regularize the tail
            nsum_tail = nsum // 10
            lowpassed_template[:nsum_tail] = np.mean(template[:nsum_tail])
            lowpassed_template[-nsum_tail:] = np.mean(template[-nsum_tail:])
            lowpassed.append(lowpassed_template)
        lowpassed = np.vstack(lowpassed)
        return lowpassed


    @function_timer
    def _load_dark_templates(self, ob):
        """ Try loading dark bolometer templates
        """
        log = Logger.get()
        comm = ob.comm.comm_group

        # See if the templates were already cached

        if self.cache_dir is not None:
            fname_cache = os.path.join(
                self.cache_dir,
                f"dark_templates.{ob.name}.pck",
            )
            if os.path.isfile(fname_cache):
                if comm is None or comm.rank == 0:
                    with open(fname_cache, "rb") as f:
                        templates = pickle.load(f)
                    log.debug(f"Loaded dark templates from {fname_cache}")
                else:
                    templates = None
                self._share_templates(ob, templates)
                return ob.shared[self.key]

        return None

    @function_timer
    def _share_templates(self, ob, templates):
        """ Put the dark templates in shared memory

        """
        comm = ob.comm.comm_group

        if templates is None:
            ntemplate, nsample = None, None
        else:
            ntemplate, nsample = templates.shape

        if comm is not None:
            ntemplate = comm.bcast(ntemplate)
            nsample = comm.bcast(nsample)

        ob.shared.create_column(
            self.key,
            shape=(nsample, ntemplate),
            dtype=np.float64,
        )
        if ob.comm.group_rank == 0:
            temp_trans = templates.T
        else:
            temp_trans = None
        ob.shared[self.key].set(temp_trans, offset=(0, 0), fromrank=0)
        return

    @function_timer
    def _save_dark_templates(self, ob, templates):
        """ Write the dark templates in a pickle file

        """
        # See if we are caching the templates

        log = Logger.get()
        comm = ob.comm.comm_group

        if self.cache_dir is None:
            return

        if comm is None or comm.rank == 0:
            fname_cache = os.path.join(
                self.cache_dir,
                f"dark_templates.{ob.name}.pck",
            )
            with open(fname_cache, "wb") as f:
                pickle.dump(templates.data.T, f)
            log.debug(f"Wrote dark templates to {fname_cache}")

        return

    @function_timer
    def _get_dark_templates(self, ob):
        """ Construct low-passed dark bolometer templates
        """
        log = Logger.get()
        comm = ob.comm.comm_group

        # gather all dark TOD to the root process

        dark_tod = self._gather_dark_tod(ob)

        # Construct dark templates through PCA and lowpass

        if comm is None or comm.rank == 0:
            templates = self._derive_templates(dark_tod)
        else:
            templates = None

        self._share_templates(ob, templates)

        return ob.shared[self.key]

    @function_timer
    def _project_dark_templates(self, ob, templates):
        """ Project the provided templates out of every optical detector

        """
        if self.shared_flags is not None:
            common_flags = ob.shared[self.shared_flags].data & self.shared_flag_mask
        else:
            common_flags = np.zeros(ob.n_local_samples, dtype=np.uint8)

        fp = ob.telescope.focalplane
        last_good = None
        for det in ob.select_local_detectors(flagmask=self.det_mask):
            if fp[det]["det_info:wafer:type"] == "DARK":
                continue
            tod = ob.detdata[self.det_data][det]
            if self.det_flags is not None:
                det_flags = ob.detdata[self.det_flags][det] & self.det_flag_mask
            else:
                det_flags = np.zeros(ob.n_local_samples, dtype=np.uint8)
            good = np.logical_and(common_flags == 0, det_flags == 0)
            intervals = ob.intervals[self.view]
            if last_good is None or np.any(good != last_good):
                recompute = True
                last_good = good
                last_covs = []
                last_templates = []
            else:
                recompute = False
            for i, ival in enumerate(intervals):
                ind = slice(ival.first, ival.last)
                good_ind = good[ind].copy()
                if recompute:
                    templates_ind = templates[ind].copy()
                    masked_templates = templates_ind[good_ind].copy()
                    invcov = np.dot(masked_templates.T, masked_templates)
                    cov = np.linalg.inv(invcov)
                    last_covs.append(cov)
                    last_templates.append(masked_templates)
                else:
                    cov = last_covs[i]
                    masked_templates = last_templates[i]
                proj = np.dot(masked_templates.T, tod[ind][good_ind])
                coeff = np.dot(cov, proj)
                tod[ind] -= np.dot(coeff, templates_ind.T)

        return

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        timer = Timer()
        comm = data.comm.comm_world
        gcomm = data.comm.comm_group
        timer.start()

        if self.cache_dir is not None:
            if comm is None or comm.rank == 0:
                os.makedirs(self.cache_dir, exist_ok=True)

        for ob in data.obs:
            if ob.dist.comm_row_size != 1:
                raise RuntimeError(
                    "DarkTemplate only works with observations distributed "
                    "by detector"
                )
            templates = self._load_dark_templates(ob)
            if templates is None:
                # No precomputed templates found
                templates = self._get_dark_templates(ob)
                self._save_dark_templates(ob, templates)
            self._project_dark_templates(ob, templates)

        return

    @function_timer
    def _finalize(self, data, **kwargs):
        pass

    def _requires(self):
        req = {
            "meta": list(),
            "shared": list(),
            "detdata": [self.det_data],
            "intervals": list(),
        }
        if self.view is not None:
            req["intervals"].append(self.view)
        if self.shared_flags is not None:
            req["shared"].append(self.shared_flags)
        if self.det_flags is not None:
            req["detdata"].append(self.det_flags)
        return req

    def _provides(self):
        prov = {"detdata": list()}
        return prov

    def _accelerators(self):
        return list()



"""TOD loader functions, within context system, for HDF5 format output
from TOAST (especially TOAST 3 circa 2023).

"""

import so3g
import numpy as np

from .. import core, coords


def load_toast_h5_obs(db, obs_id, dets=None, samples=None, prefix=None,
                      no_signal=None,
                      **kwargs):
    """Obsloader function for TOAST HDF5 output TODs.  Based on output
    from toast_so_sim.py, SSO sims, toast 3.0.0a9.

    See API template, `sotodlib.core.context.obsloader_template`, for
    details.

    """
    if prefix is None:
        prefix = db.prefix
        if prefix is None:
            prefix = './'

    # Get a ResultSet<[dets:detset,dets:readout_id], N rows>
    props = db.get_det_table(obs_id)
    if dets is None:
        detsets = props.subset(keys=['dets:detset']).distinct()['dets:detset']
    elif len(dets) == 0:
        props = props.subset(rows=[0])
        detsets = props['dets:detset']
    else:
        # Cross-match.
        vals, i0, i1 = core.util.get_coindices(props['dets:readout_id'], dets)
        if len(i1) != len(dets):
            raise ValueError('Some dets not found.')
        # Implicated detsets?
        detsets = props.subset(rows=i0, keys=['dets:detset']).distinct()['dets:detset']

    files_by_detset = db.get_files(obs_id, detsets=detsets)

    components = []
    for detset, files in files_by_detset.items():
        assert(len(files) == 1)
        dets = props.subset(rows=(props['dets:detset'] == detset))['dets:readout_id']
        components.append(
            load_toast_h5_file(
                files[0][0], dets=dets, no_signal=no_signal, samples=samples))
    if len(components) > 1:
        msg = "This loader only supports loading detectors from the same wafer"
        raise RuntimeError(msg)
    return components[0]

def load_toast_h5_dichroic_hack(db, obs_id, dets=None, samples=None, prefix=None,
                                no_signal=None,
                                **kwargs):
    """Obsloader function for TOAST HDF5 output TODs, in the specific case
    that each detset contains detectors for two bands and those
    detectors are distributed into two separate HDF5 files.

    See API template, `sotodlib.core.context.obsloader_template`, for
    details.

    """
    if prefix is None:
        prefix = db.prefix
        if prefix is None:
            prefix = './'

    # Get a ResultSet<[dets:detset,dets:readout_id], N rows>
    props = db.get_det_table(obs_id)
    if dets is None:
        detsets = props.subset(keys=['dets:detset']).distinct()['dets:detset']
    elif len(dets) == 0:
        props = props.subset(rows=[0])
        detsets = props['dets:detset']
    else:
        # Cross-match.
        vals, i0, i1 = core.util.get_coindices(props['dets:readout_id'], dets)
        if len(i1) != len(dets):
            raise ValueError('Some dets not found.')
        # Implicated detsets?
        detsets = props.subset(rows=i0, keys=['dets:detset']).distinct()['dets:detset']

    # Filter dets by band ...
    band_dets = {b: [] for b in ['f090', 'f150']}
    for d in props['dets:readout_id']:
        for b, dest in band_dets.items():
            if b in d:
                dest.append(d)
                break

    bands_here = sorted([b for b, d in band_dets.items() if len(d)])
    files_by_detset = db.get_files(obs_id, detsets=detsets)
    base_band = [b for b in bands_here if b in list(files_by_detset.items())[0][1][0][0]][0]

    components = []
    for detset, files in files_by_detset.items():
        assert(len(files) == 1)
        detset_dets = props.subset(rows=(props['dets:detset'] == detset))['dets:readout_id']
        for band in bands_here:
            band_file = files[0][0].replace(base_band, band)
            _dets = [d for d in detset_dets if d in band_dets[band]]
            components.append(
                load_toast_h5_file(
                    band_file, dets=_dets, no_signal=no_signal, samples=samples))

    if len(components) > 1:
        return components[0].concatenate(components, axis='dets', other_fields='first')

    return components[0]

def load_toast_h5_file(filename, dets=None, samples=None, no_signal=False,
                       enhance=False):
    """Reads data from a single HDF5 file.  Returns an AxisManager with
    standard SO field names.  Supports load_toast_h5_obs.

    If enhance=True, then certain metadata from the file are also
    processed and included: "focal_plane".

    """
    import h5py

    with h5py.File(filename, 'r') as f:
        dets_in_file = [v['name'].decode('ascii') for v in f['instrument']['focalplane']]

        if dets is None:
            dets = dets_in_file
            dets_mode = 'all'
        elif len(dets) == 0:
            dets_mode = 'none'
        else:
            #  Allow that dets might be identical to dets_in_file...
            vals, i0, i1 = core.util.get_coindices(dets_in_file, dets)
            if len(i0) != len(dets) or np.any(i1 != i0):
                raise ValueError('dets should be None or [].')
            dets = dets_in_file
            dets_mode = 'all'

        count = f['detdata']['signal'].shape[1]

        aman = core.AxisManager(
            core.LabelAxis('dets', dets),
            core.OffsetAxis('samps', count, 0),
        )

        aman.wrap_new('timestamps', ('samps', ), dtype='float64')

        bman = core.AxisManager(aman.samps.copy())
        bman.wrap('az', f['shared']['azimuth'], [(0, 'samps')])
        bman.wrap('el', f['shared']['elevation'], [(0, 'samps')])
        bman.wrap('roll', 0*bman.az, [(0, 'samps')])
        aman.wrap('boresight', bman)

        aman.timestamps[:] = f['shared']['times']
        if dets_mode == 'all':
            if not no_signal:
                aman.wrap_new('signal', ('dets', 'samps'), dtype='float32')
                aman.signal[:] = f['detdata']['signal']
        else:
            assert(len(dets) == 0)

        if 'hwp_angle' in f['shared']:
            aman.wrap('hwp_angle', f['shared']['hwp_angle'][()], [(0, 'samps')])

        # For the LAT corotator angle / SAT boresight rotation --
        # store each one, but also process them into a boresight
        # "roll".

        if 'corotator_angle' in f['shared']:
            a = f['shared']['corotator_angle'][()]
            aman.wrap('corotator_angle', a, [(0, 'samps')])

            # When uncorrected by the corotator, increasing elevation
            # causes footprint on sky to rotate clockwise.
            aman.boresight.roll[:] = (aman.boresight.el - np.radians(60) - a)

        if 'boresight_angle' in f['shared']:
            a = f['shared']['boresight_angle'][()]
            aman.wrap('boresight_angle', a, [(0, 'samps')])

            # Positive boresight angle corresponds to
            # counter-clockwise rotation of projection on sky, which
            # is a negative roll angle.
            aman.boresight.roll[:] = -a

        # Save the full toast fp info -- lots of interesting stuff in there.
        aman.wrap('toast_focalplane', f['instrument']['focalplane'], [(0, 'dets')])

    if enhance:
        # Write out equivalent sotodlib det pos.
        q = coords.ScalarLastQuat(aman.toast_focalplane['quat']).to_g3()
        xi, eta, gamma = so3g.proj.quat.decompose_xieta(q)
        fp = core.AxisManager(aman.dets)
        for key, value in zip(['xi', 'eta', 'gamma'], [xi, eta, gamma]):
            fp.wrap_new(key, shape=('dets', ))[:] = value
        aman.wrap('focal_plane', fp)

    if samples is not None:
        aman.restrict('samps', samples)

    return aman


core.OBSLOADER_REGISTRY['toast3-hdf-dichroic-hack'] = load_toast_h5_obs



# Copyright (c) 2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check functionality of the wiregrid simulation

"""

import unittest
import numpy as np

import astropy.units as u

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import toast
    from toast.observation import default_values as defaults
    toast_available = True
except ImportError as e:
    toast_available = False

from ._helpers import calibration_schedule, close_data_and_comm, simulation_test_data


class SimWireGridTest(unittest.TestCase):
    def test_sim_wiregrid(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="SAT1",
            wafer_slot=None,
            bands="SAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=16,
            cal_schedule=True,
        )

        pointing = toast.ops.PointingDetectorSimple(
            name="det_pointing_azel",
            quats="quats_azel",
            boresight=defaults.boresight_azel,
        )
        weights = toast.ops.StokesWeights(
            name="weights_azel",
            weights="weights_azel",
            mode="IQU",
            detector_pointing=pointing,
        )

        so_ops.SimWireGrid(
            name="sim_wiregrid",
            detector_pointing=pointing,
            detector_weights=weights,
            wiregrid_angular_speed=10.0*u.degree/u.second,
            wiregrid_angular_acceleration=0.1*u.degree/u.second**2,
        ).apply(data)

        close_data_and_comm(data)

if __name__ == '__main__':
    unittest.main()



import os
import unittest
import json
import shutil

from sotodlib.core import resources

ODD_NAME = 'planet47-ephemeris.txt'


class TestCoreResources(unittest.TestCase):

    def test_get_local_file(self):

        # This url should not be accessed, as long as download=False
        resources.RESOURCE_DEFAULTS[ODD_NAME] = \
            'ftp://simonsobs.org/sotodlib-ci.txt'

        os.environ["SOTODLIB_RESOURCES"] = json.dumps(
            {"someotherfile": "file://somepath/otherfile"}
        )

        # Use a fake file for this, or else it might find a
        # user-cached copy.
        t = resources.get_local_file(ODD_NAME, cache=False, download=False)
        expected_path = "/tmp/" + ODD_NAME
        self.assertEqual(expected_path, t)

        os.environ["SOTODLIB_RESOURCES"] = json.dumps(
            {ODD_NAME: "file://somepath/de421.bsp"}
        )
        t = resources.get_local_file(ODD_NAME)
        self.assertEqual("somepath/de421.bsp", t)

        with self.assertRaises(RuntimeError):
            resources.get_local_file("doesnotexist.file")



import logging
import os
from dataclasses import InitVar, dataclass, field
from functools import cached_property, partial
from typing import Dict, List, Optional

import h5py
import matplotlib.pyplot as plt
import megham.transform as mt
import megham.utils as mu
import numpy as np
from numpy.typing import NDArray
from scipy.stats import binned_statistic
from sotodlib.coords import optics as op
from sotodlib.core import metadata
from sotodlib.io.metadata import read_dataset, write_dataset

logger = logging.getLogger("finalize_focal_plane")
plt.style.use("tableau-colorblind10")


def _add_attrs(dset, attrs):
    for k, v in attrs.items():
        if v is None:
            logger.warning("Not adding attribute %s because it is None", k)
            continue
        dset.attrs[k] = v


@dataclass
class Transform:
    shift: NDArray[np.floating]  # (ndim,)
    affine: NDArray[np.floating]  # (ndim, ndim)
    scale: NDArray[np.floating] = field(init=False)  # (ndim,)
    shear: float = field(init=False)
    rot: float = field(init=False)

    def __post_init__(self):
        self.decompose()

    @classmethod
    def identity(cls):
        return Transform(np.zeros(3), np.eye(3))

    @classmethod
    def from_split(cls, shift, xieta_affine, gamma_scale):
        affine = np.eye(len(xieta_affine) + 1)
        affine[: len(xieta_affine), : len(xieta_affine)] = xieta_affine
        affine[-1, -1] = gamma_scale

        return Transform(shift, affine)

    def decompose(self):
        xieta_affine = self.affine[:2, :2]
        gamma_scale = self.affine[-1, -1]
        scale, shear, rot = mt.decompose_affine(xieta_affine)
        self.scale = np.array((*scale, gamma_scale))
        self.shear = shear.item()
        self.rot = mt.decompose_rotation(rot)[-1]

    def save(self, f, path, append=""):
        if path not in f:
            f.create_group(path)
        _add_attrs(
            f[path],
            {
                f"shift{append}": self.shift,
                f"scale{append}": self.scale,
                f"shear{append}": self.shear,
                f"rot{append}": self.rot,
                f"affine{append}": self.affine,
            },
        )

    @classmethod
    def load(cls, group, append=""):
        shift = group.attrs[f"shift{append}"]
        affine = group.attrs[f"affine{append}"]

        return Transform(shift, affine)


@dataclass
class Template:
    det_ids: NDArray[np.str_]  # (ndet,)
    fp: NDArray[np.floating]  # (ndet, ndim)
    optical: NDArray[np.bool_]  # (ndet,)
    pointing_cfg: InitVar[Dict]
    center: NDArray[np.floating] = field(init=False)  # (1, ndim)
    spacing: NDArray[np.floating] = field(init=False)  # (ndim,)
    bandpass: NDArray[np.str_] = field(init=False)
    pol: NDArray[np.str_] = field(init=False)
    rhombus: NDArray[np.str_] = field(init=False)

    def __post_init__(self, pointing_cfg):
        self.center = np.array(
            op.get_focal_plane(None, x=0, y=0, pol=0, **pointing_cfg)
        ).T
        xieta_spacing = mu.estimate_spacing(self.fp[self.optical, :2])
        # For gamma rather than the spacing in real space we want the difference between bins
        # This is a rough estimate but good enough for us
        gamma_spacing = np.percentile(np.diff(np.sort(self.fp[2])), 99.9)
        self.spacing = np.array([xieta_spacing, xieta_spacing, gamma_spacing])

        self.bandpass = np.zeros_like(self.det_ids)
        self.pol = np.zeros_like(self.det_ids)
        self.rhombus = np.zeros_like(self.det_ids)

    def add_wafer_info(self, aman, template_msk):
        self.__dict__.pop("id_strs", None)
        self.__dict__.pop("valid_ids", None)
        if not np.all(
            np.isin(["bandpass", "pol", "rhombus"], list(aman.det_info.wafer.keys()))
        ):
            logger.warning(
                "det_info.wafer seems to be missing metadata? Safe to ignore in rset mode."
            )
            return
        mapping = np.argsort(np.argsort(self.det_ids[template_msk]))
        srt = np.argsort(aman.det_info.det_id)
        self.bandpass[template_msk] = aman.det_info.wafer.bandpass[srt][mapping]
        self.pol[template_msk] = aman.det_info.wafer.pol[srt][mapping]
        self.rhombus[template_msk] = aman.det_info.wafer.rhombus[srt][mapping]

    @cached_property
    def id_strs(self):
        return np.array([bp + "_" + pol for bp, pol in zip(self.bandpass, self.pol)])

    @cached_property
    def valid_ids(self):
        ids = np.unique(self.id_strs)
        valid = ~(np.char.startswith(ids, "_") * np.char.endswith(ids, "_"))
        return ids[valid]


@dataclass
class FocalPlane:
    stream_id: str
    wafer_slot: str
    det_ids: NDArray[np.str_]  # (ndet,)
    id_strs: NDArray[np.str_]  # (ndet,)
    avg_fp: NDArray[np.floating]  # (ndim, ndet)
    weights: NDArray[np.floating]  # (ndet,)
    transformed: NDArray[np.floating]  # (ndet, ndim)
    center: NDArray[np.floating]  # (1, ndim)
    center_transformed: NDArray[np.floating]  # (1, ndim)
    n_point: NDArray[np.int_]  # (ndet,)
    n_gamma: NDArray[np.int_]  # (ndet,)
    have_gamma: bool = field(default=False)
    template: Optional[Template] = field(default=None)
    transform: Transform = field(default_factory=Transform.identity)
    transform_nocm: Transform = field(default_factory=Transform.identity)
    config: str = field(default="")
    with_cm: bool = field(default=True)
    full_fp: Optional[NDArray[np.floating]] = field(
        init=False, default=None
    )  # (ndet, ndim, n_aman)
    tot_weight: Optional[NDArray[np.floating]] = field(
        init=False, default=None
    )  # (ndet,)

    def __post_init__(self):
        if self.template is not None and not np.all(
            np.isclose(self.center, self.template.center)
        ):
            raise ValueError("Focalplane center does not match template")

    @property
    def template_fp(self):
        if self.template is not None:
            return self.template.fp
        logger.warning(
            "No template in focal plane %s, reconstructing from transformed. This will be wrong with_cm is set wrong!",
            self.stream_id,
        )
        transform = self.transform_nocm
        if self.with_cm:
            transform = self.transform
        return (self.transformed - transform.shift) @ np.linalg.inv(transform.affine)

    @property
    def template_center(self):
        if self.template is not None:
            return self.template.center
        logger.warning(
            "No template in focal plane %s, reconstructing from transformed. This will be wrong with_cm is set wrong!",
            self.stream_id,
        )
        transform = self.transform_nocm
        if self.with_cm:
            transform = self.transform
        return (self.center_transformed - transform.shift) @ np.linalg.inv(
            transform.affine
        )

    @property
    def diff(self):
        return self.avg_fp - self.transformed

    @property
    def isfinite(self):
        return np.isfinite(self.diff[:, 0])

    @property
    def dist(self):
        if self.have_gamma:
            return np.linalg.norm(self.diff, axis=1)
        return np.linalg.norm(self.diff[:, :2], axis=1)

    @property
    def padded(self):
        return self.tot_weight is None

    @classmethod
    def empty(cls, template, stream_id, wafer_slot, n_aman, config=""):
        if template is None:
            raise TypeError("template must be an instance of Template, not None")
        full_fp = np.full(template.fp.shape + (n_aman,), np.nan)
        tot_weight = np.zeros((len(template.det_ids), 2))
        avg_fp = np.full_like(template.fp, np.nan)
        weight = np.zeros((len(template.det_ids), 2))
        transformed = template.fp.copy()
        center = template.center.copy()
        center_transformed = template.center.copy()
        n_point = np.zeros_like(template.det_ids, dtype=int)
        n_gamma = np.zeros_like(template.det_ids, dtype=int)

        fp = FocalPlane(
            stream_id,
            wafer_slot,
            template.det_ids,
            template.id_strs,
            avg_fp,
            weight,
            transformed,
            center,
            center_transformed,
            n_point,
            n_gamma,
            template=template,
            config=config,
        )
        fp.full_fp = full_fp
        fp.tot_weight = tot_weight

        return fp

    def map_by_det_id(self, aman):
        if self.template is not None and not np.array_equal(
            self.det_ids, self.template.det_ids
        ):
            raise ValueError("det_ids don't match template, mapping is not valid!")
        _, msk, template_msk = np.intersect1d(
            aman.det_info.det_id, self.det_ids, return_indices=True
        )
        if len(msk) != aman.dets.count:
            logger.warning("There are matched dets not found in the focal plane")
        mapping = np.argsort(np.argsort(self.det_ids[template_msk]))
        srt = np.argsort(aman.det_info.det_id[msk])
        xi = aman.pointing.xi[msk][srt][mapping]
        eta = aman.pointing.eta[msk][srt][mapping]
        r2 = np.nan + np.zeros_like(eta)
        if "R2" in aman.pointing:
            r2 = aman.pointing.R2[msk][srt][mapping]
        if "polarization" in aman:
            # name of field just a placeholder for now
            gamma = aman.polarization.polang[msk][srt][mapping]
        elif "gamma" in aman.pointing:
            gamma = aman.pointing.gamma[msk][srt][mapping]
        else:
            gamma = np.full(len(xi), np.nan)
        fp = np.column_stack((xi, eta, gamma))
        return fp, r2, template_msk

    def add_fp(self, i, fp, weights, template_msk):
        if self.full_fp is None or self.tot_weight is None:
            raise ValueError("full_fp or tot_weight not initialized")
        self.full_fp[template_msk, :, i] = fp * weights[:, 0][..., None]
        weights = np.nan_to_num(weights)
        self.tot_weight[template_msk] += weights

    def save(self, f, db_info, group):
        logger.info("Saving %s", self.stream_id)
        ndets = len(self.det_ids)
        outdt = [
            ("dets:det_id", self.det_ids.dtype),
            ("xi", np.float32),
            ("eta", np.float32),
            ("gamma", np.float32),
            ("padded", np.bool_),
        ]
        fpout = np.fromiter(
            zip(
                self.det_ids,
                *(self.transformed.T),
                np.ones(len(self.det_ids), dtype=bool) * (self.padded),
            ),
            dtype=outdt,
            count=ndets,
        )
        write_dataset(
            metadata.ResultSet.from_friend(fpout),
            f,
            f"{group}/focal_plane",
            overwrite=True,
        )
        _add_attrs(
            f[f"{group}/focal_plane"],
            {"wafer_slot": str(self.wafer_slot), "measured_gamma": self.have_gamma},
        )
        entry = {"dets:stream_id": self.stream_id, "dataset": f"{group}/focal_plane"}
        entry.update(db_info[1])
        db_info[0].add_entry(entry, filename=os.path.basename(f.filename), replace=True)

        outdt_full = [
            ("dets:det_id", self.det_ids.dtype),
            ("id_strs", self.id_strs.dtype),
            ("xi_t", np.float32),
            ("eta_t", np.float32),
            ("gamma_t", np.float32),
            ("xi_m", np.float32),
            ("eta_m", np.float32),
            ("gamma_m", np.float32),
            ("weights", np.float32),
            ("r2", np.float32),
            ("n_point", np.int8),
            ("n_gamma", np.int8),
        ]
        fpfullout = np.fromiter(
            zip(
                self.det_ids,
                self.id_strs,
                *(self.transformed.T),
                *(self.avg_fp.T),
                *(self.weights.T),
                self.n_point,
                self.n_gamma,
            ),
            dtype=outdt_full,
            count=ndets,
        )
        write_dataset(
            metadata.ResultSet.from_friend(fpfullout),
            f,
            f"{group}/focal_plane_full",
            overwrite=True,
        )

        self.transform.save(f, f"{group}/transform")
        self.transform_nocm.save(f, f"{group}/transform", "_nocm")
        _add_attrs(
            f[f"{group}"],
            {
                "fit_centers": self.center_transformed,
                "template_centers": self.center,
                "config": self.config,
                "with_cm": self.with_cm,
                "tot_weight": self.tot_weight,
            },
        )

    @classmethod
    def load(cls, group, include_cm=None):
        stream_id = group.name.split("/")[-1]
        fp_full = read_dataset(group.file, f"{group.name}/focal_plane_full")
        if fp_full.keys is None:
            raise ValueError("fp_full somehow has no keys")
        det_ids = fp_full["dets:det_id"]
        if "id_strs" in fp_full.keys:
            id_strs = fp_full["id_strs"]
        else:
            logger.warning("No id strings found in this focal plane, may be old.")
            id_strs = np.zeros(len(fp_full), str)
        avg_fp = np.column_stack(
            (
                np.array(fp_full["xi_m"]),
                np.array(fp_full["eta_m"]),
                np.array(fp_full["gamma_m"]),
            )
        )
        # For backwards compatibility
        weights = np.array(fp_full["weights"])
        if "r2" in fp_full.keys:
            weights = np.column_stack((weights, np.array(fp_full["r2"])))
        transformed = np.column_stack(
            (
                np.array(fp_full["xi_t"]),
                np.array(fp_full["eta_t"]),
                np.array(fp_full["gamma_t"]),
            )
        )
        center = group.attrs["template_centers"]
        center_transformed = group.attrs["fit_centers"]
        n_point = fp_full["n_point"]
        n_gamma = fp_full["n_gamma"]
        have_gamma = group["focal_plane"].attrs["measured_gamma"]
        template = None
        transform = Transform.load(group["transform"])
        transform_nocm = Transform.load(group["transform"], "_nocm")
        if "wafer_slot" in group["focal_plane"].attrs:
            wafer_slot = group["focal_plane"].attrs["wafer_slot"]
        else:
            logger.warning("No wafer slot found in this focal plane, may be old.")
            wafer_slot = "ws?"
        config = group.attrs.get("config", "")
        if config == "":
            logger.warning("No config found in this focal plane, may be old.")
        if "with_cm" in group.attrs:
            with_cm = group.attrs["with_cm"]
        else:
            with_cm = True
            logger.warning(
                "No with_cm found in this focal plane, may be old. Attempting to set based on receiver!"
            )
            if include_cm is None:
                logger.warning("\tinclude_cm not set! Using True for with_cm")
            else:
                with_cm = include_cm

        fp = FocalPlane(
            stream_id,
            wafer_slot,
            np.array(det_ids),
            np.array(id_strs),
            avg_fp,
            np.array(weights),
            transformed,
            center,
            center_transformed,
            np.array(n_point),
            np.array(n_gamma),
            have_gamma,
            template,
            transform,
            transform_nocm,
            config,
            with_cm,
        )
        fp.tot_weight = group.attrs.get("tot_weight", None)

        return fp


@dataclass
class OpticsTube:
    name: str
    center: NDArray[np.floating]
    transform: Transform = field(default_factory=Transform.identity)
    transform_fullcm: Transform = field(default_factory=Transform.identity)
    focal_planes: List[FocalPlane] = field(default_factory=list)
    center_transformed: NDArray[np.floating] = field(init=False)

    def __post_init__(self):
        self.center_transformed = mt.apply_transform(
            self.center, self.transform_fullcm.affine, self.transform_fullcm.shift
        )

    def __setattr__(self, name, val):
        self.__dict__[name] = val
        if name == "focal_planes":
            self.__dict__.pop("fp_dict", None)

    @property
    def num_fps(self):
        fps = [fp for fp in self.focal_planes if fp.tot_weight is not None]
        return len(fps)

    @classmethod
    def from_pointing_cfg(cls, pointing_cfg):
        name = pointing_cfg["tube_slot"]
        telescope_flavor = pointing_cfg["telescope_flavor"].upper()
        if telescope_flavor not in ["LAT", "SAT"]:
            raise ValueError("Telescope should be LAT or SAT")

        if telescope_flavor == "LAT":
            if pointing_cfg["zemax_path"] is None:
                raise ValueError("Must provide zemax_path for LAT")
            xi, eta, gamma = op.LAT_focal_plane(
                None,
                pointing_cfg["zemax_path"],
                x=0,
                y=0,
                pol=0,
                roll=pointing_cfg.get("roll", 0),
                tube_slot=name,
            )
        else:
            xi, eta, gamma = op.SAT_focal_plane(
                None,
                x=0,
                y=0,
                pol=0,
                roll=pointing_cfg.get("roll", 0),
                mapping_data=pointing_cfg.get("mapping_data", None),
            )
        center = np.array((xi, eta, gamma)).reshape((1, 3))

        return OpticsTube(name, center)

    def delete_fp(self, stream_id):
        self.focal_planes = [
            fp for fp in self.focal_planes if fp.stream_id != stream_id
        ]

    @cached_property
    def fp_dict(self):
        return {fp.stream_id: fp for fp in self.focal_planes}

    def save(self, f, db_info, group="/"):
        g = f[group]
        g.create_group(self.name)
        _add_attrs(
            g[self.name],
            {"center": self.center, "center_transformed": self.center_transformed},
        )
        tr_path = os.path.join(group, self.name, "transform")
        self.transform.save(f, tr_path)
        self.transform_fullcm.save(f, tr_path, "_fullcm")
        for focal_plane in self.focal_planes:
            focal_plane.save(
                f, db_info, os.path.join(group, self.name, focal_plane.stream_id)
            )

    @classmethod
    def load(cls, group, include_cm=None):
        name = group.name.split("/")[-1]
        center = group.attrs["center"]
        transform = Transform.load(group["transform"])
        transform_fullcm = Transform.load(group["transform"], "_fullcm")
        fps = [
            FocalPlane.load(group[grp], include_cm)
            for grp in group.keys()
            if "transform" not in grp
        ]

        return OpticsTube(name, center, transform, transform_fullcm, fps)


@dataclass
class Receiver:
    optics_tubes: List[OpticsTube] = field(default_factory=list)
    center: NDArray[np.floating] = field(
        default_factory=partial(np.zeros, shape=(1, 3))
    )
    include_cm: bool = field(default=False)
    valid_ids: List[str] = field(default_factory=list)
    transform: Transform = field(default_factory=Transform.identity)
    center_transformed: NDArray[np.floating] = field(init=False)

    def __post_init__(self):
        self.center_transformed = mt.apply_transform(
            self.center, self.transform.affine, self.transform.shift
        )

    def __setattr__(self, name, val):
        self.__dict__[name] = val
        if name == "optics_tubes":
            self.__dict__.pop("ot_dict", None)

    def delete_ot(self, name):
        self.focal_planes = [ot for ot in self.optics_tubes if ot.name != name]

    def save(self, f, db_info, group="/"):
        _add_attrs(
            f[group],
            {
                "center": self.center,
                "center_transformed": self.center_transformed,
                "include_cm": self.include_cm,
                "valid_ids": self.valid_ids,
            },
        )
        self.transform.save(f, os.path.join(group, "transform"))
        for ot in self.optics_tubes:
            ot.save(f, db_info, group)

    @classmethod
    def load(cls, f, group="/"):
        center = f[group].attrs["center"]
        include_cm = f[group].attrs["include_cm"]
        valid_ids = f[group].attrs.get("valid_ids", [])
        if len(valid_ids) == 0:
            logger.warning("valid_ids list is empty! May be an old Receiver object!")
        transform = Transform.load(f[group]["transform"])
        ots = [
            OpticsTube.load(f[group][grp], include_cm)
            for grp in f[group].keys()
            if "transform" not in grp
        ]

        return Receiver(ots, center, include_cm, list(valid_ids), transform)

    @classmethod
    def load_file(cls, path):
        with h5py.File(path, "r") as f:
            # Check if its an old file:
            if "transform" in f.keys():
                return {"": cls.load(f)}
            return {grp: cls.load(f, grp) for grp in f.keys()}

    @property
    def focal_planes(self):
        fps = []
        for ot in self.optics_tubes:
            fps += ot.focal_planes
        return fps

    @cached_property
    def ot_dict(self):
        return {ot.name: ot for ot in self.optics_tubes}

    @property
    def lims(self):
        xmax = np.max([np.nanmax(fp.transformed[:, 0]) for fp in self.focal_planes])
        xmin = np.min([np.nanmin(fp.transformed[:, 0]) for fp in self.focal_planes])
        ymax = np.max([np.nanmax(fp.transformed[:, 1]) for fp in self.focal_planes])
        ymin = np.min([np.nanmin(fp.transformed[:, 1]) for fp in self.focal_planes])
        return (xmin, xmax), (ymin, ymax)

    @property
    def fp_valid_ids(self):
        valid_ids = []
        for fp in self.focal_planes:
            if fp.template is None:
                continue
            valid_ids += list(fp.template.valid_ids)
        valid_ids = np.unique(valid_ids)
        return valid_ids.tolist()


# Plotting Functions
def plot_ufm(focal_plane, plot_dir):
    nominal = focal_plane.template_fp
    measured = focal_plane.avg_fp
    transformed = focal_plane.transformed
    fig, (ax1, ax2) = plt.subplots(1, 2, constrained_layout=True)
    # Plot pointing
    ax1.scatter(
        nominal[:, 0],
        nominal[:, 1],
        alpha=0.4,
        color="blue",
        label="nominal",
        marker="P",
    )
    ax1.scatter(
        transformed[:, 0],
        transformed[:, 1],
        alpha=0.4,
        color="black",
        label="transformed",
        marker="X",
    )
    ax1.scatter(measured[:, 0], measured[:, 1], alpha=0.4, color="orange", label="fit")
    ax1.set_xlabel("Xi (rad)")
    ax1.set_ylabel("Eta (rad)")
    ax1.set_aspect("equal")
    ax1.legend()

    # Histogram of differences
    dist = focal_plane.dist[focal_plane.isfinite] * 180 * 60 * 60 / np.pi
    dist_thresh = np.percentile(dist, 97)
    bins = max(int(len(dist) / 20), 10)
    ax2.hist(dist[dist < dist_thresh], bins=bins)
    ax2.set_xlabel("Residual (arcseconds)")

    fig.suptitle(f"{focal_plane.stream_id}")
    fig.set_size_inches(2 * fig.get_size_inches())
    if plot_dir is None:
        plt.show()
    else:
        os.makedirs(plot_dir, exist_ok=True)
        plt.savefig(
            os.path.join(plot_dir, f"{focal_plane.stream_id}.png"), bbox_inches="tight"
        )
        plt.close()


def plot_ot(ot, plot_dir):
    fig, (ax1, ax2) = plt.subplots(1, 2, constrained_layout=True)
    dists = [
        fp.dist[fp.isfinite] * 180 * 60 * 60 / np.pi
        for fp in ot.focal_planes
        if fp.tot_weight is not None
    ]
    xis = [
        fp.transformed[fp.isfinite, 0]
        for fp in ot.focal_planes
        if fp.tot_weight is not None
    ]
    etas = [
        fp.transformed[fp.isfinite, 1]
        for fp in ot.focal_planes
        if fp.tot_weight is not None
    ]

    # Plot the radial dist
    r = np.sqrt(np.hstack(xis) ** 2 + np.hstack(etas) ** 2)
    dist = np.hstack(dists)
    dist_avg, edges, _ = binned_statistic(r, dist, "median", bins=50)
    r_bins = (edges[:-1] + edges[1:]) / 2
    max_dist = np.max(dist)

    ax1.scatter(r, dist, alpha=0.1)
    ax1.plot(r_bins, dist_avg, color="black")
    ax1.set_ylim((None, np.percentile(dist, 95)))
    ax1.set_xlabel("Radius (rad)")
    ax2.set_ylabel("Residual (arcseconds)")

    # Plot a heatmap
    cf = None
    for dist, xi, eta in zip(dists, xis, etas):
        cf = ax2.tricontourf(
            xi, eta, dist, levels=20, vmin=-1 * max_dist, vmax=max_dist, cmap="coolwarm"
        )
    if cf is not None:
        fig.colorbar(cf, ax=ax2, label="arcseconds")
    ax2.set_aspect("equal")
    ax2.set_xlabel("Xi (rad)")
    ax2.set_ylabel("Eta (rad)")
    fig.suptitle(f"{ot.name}")

    fig.set_size_inches(2 * fig.get_size_inches())
    if plot_dir is None:
        plt.show()
    else:
        os.makedirs(plot_dir, exist_ok=True)
        plt.savefig(os.path.join(plot_dir, f"{ot.name}.png"), bbox_inches="tight")
        plt.close()


def plot_by_gamma(focal_plane, plot_dir):
    fig, axs = plt.subplots(1, 3, sharey=True, constrained_layout=True)
    dist_thresh = np.percentile(focal_plane.dist[focal_plane.isfinite], 97)
    msk = (focal_plane.dist < dist_thresh) + focal_plane.isfinite
    rhombi = np.unique(focal_plane.template.rhombus)
    gammas = (focal_plane.template.fp[msk, 2] * 180 / np.pi) % 180.0
    bins = np.linspace(0, 180, 13)
    for i, name in enumerate(("xi", "eta", "gamma")):
        d = focal_plane.diff[msk, i] * 180 * 60 * 60 / np.pi
        axs[i].set_title(name)
        if np.sum(np.isfinite(d)) == 0:
            continue
        medians, *_ = binned_statistic(gammas, d, statistic="median", bins=bins)
        for rhombus in rhombi:
            rmsk = focal_plane.template.rhombus[msk] == rhombus
            if not (np.any(rmsk)):
                continue
            axs[i].scatter(gammas[rmsk], d[rmsk], alpha=0.2, label=f"Rhombus {rhombus}")
        axs[i].scatter(np.arange(7.5, 180, 15), medians, color="black")
        leg = axs[i].legend()
        for lh in leg.legend_handles:
            lh.set_alpha(1)
    axs[0].set_ylabel("Diff (arcseconds)")
    axs[1].set_xlabel("Nominal Gamma (deg)")
    fig.suptitle(f"{focal_plane.stream_id} By Gamma")
    fig.set_size_inches(2 * fig.get_size_inches())
    if plot_dir is None:
        plt.show()
    else:
        os.makedirs(plot_dir, exist_ok=True)
        plt.savefig(
            os.path.join(plot_dir, f"{focal_plane.stream_id}_by_gamma.png"),
            bbox_inches="tight",
        )
        plt.close()


def plot_receiver(receiver, plot_dir):
    valid_ids = receiver.valid_ids
    max_diff = 0
    for fp in receiver.focal_planes:
        diff = np.nanpercentile(np.abs(fp.diff), 97)
        if diff > max_diff:
            max_diff = diff
    max_diff *= 180 * 60 * 60 / np.pi
    xlims, ylims = receiver.lims

    fig, axs_all = plt.subplots(
        len(valid_ids), 3, sharex="col", sharey="row", constrained_layout=True
    )
    axs = axs_all.flat
    axs[0].set_title("Xi")
    axs[1].set_title("Eta")
    axs[2].set_title("Gamma")
    cf = None
    for i in range(len(valid_ids)):
        for j in range(3):
            axs[3 * i + j].set_aspect("equal")
            axs[3 * i + j].set_xlim(xlims)
            axs[3 * i + j].set_ylim(ylims)
        for fp in receiver.focal_planes:
            if fp.tot_weight is None:
                continue
            msk = (fp.id_strs == valid_ids[i]) * fp.isfinite
            if np.sum(msk) < 3:
                continue
            diff = fp.diff * 180 * 60 * 60 / np.pi
            cf = axs[3 * i + 0].tricontourf(
                fp.transformed[msk, 0],
                fp.transformed[msk, 1],
                diff[msk, 0],
                levels=20,
                vmin=-1 * max_diff,
                vmax=max_diff,
                cmap="coolwarm",
            )
            cf = axs[3 * i + 1].tricontourf(
                fp.transformed[msk, 0],
                fp.transformed[msk, 1],
                diff[msk, 1],
                levels=20,
                vmin=-1 * max_diff,
                vmax=max_diff,
                cmap="coolwarm",
            )
            if fp.have_gamma:
                cf = axs[3 * i + 2].tricontourf(
                    fp.transformed[msk, 0],
                    fp.transformed[msk, 1],
                    diff[msk, 2],
                    levels=20,
                    vmin=-1 * max_diff,
                    vmax=max_diff,
                    cmap="coolwarm",
                )
        axs[3 * i + 0].set_ylabel(f"{valid_ids[i]}\nEta (rad)")
    if cf is not None:
        fig.colorbar(cf, ax=axs_all.ravel().tolist(), label="arcsecs")
    axs[-3].set_xlabel("Xi (rad)")
    axs[-2].set_xlabel("Xi (rad)")
    axs[-1].set_xlabel("Xi (rad)")
    fig.suptitle("Full Receiver Residuals")
    fig.set_size_inches(2 * fig.get_size_inches())
    if plot_dir is None:
        plt.show()
    else:
        os.makedirs(plot_dir, exist_ok=True)
        plt.savefig(os.path.join(plot_dir, f"receiver.png"), bbox_inches="tight")
        plt.close()



"""Functions and classes to support the loading of TOD data from files
on disk.  The routines provide the standard no-TOAST representation of
SO TOD data, and also should be leveraged to populate TOAST TOD
objects.

The basic routines here provide structures for automatically unpacking
series of G3 frames into coherent data structures.

Additional routines use an ObsFileDb (provided by the user) to
optimize disk reads for loading particular data of interest to a user.
These routines rely on the completeness and accuracy of the ObsFileDb,
and are appropriate for large archives of files that have been
pre-scanned to log their contents.

There will inevitably be a few different "file formats" out there, at
the very least some variations in the schema used on top of G3.
Presently the loader targets the pipe-s0001 "1day" TOD sim for the
LAT.

To support different loaders, we will host (here) a registry of named
loader functions.

"""

import so3g
from spt3g import core as g3core

import numpy as np

from collections import OrderedDict

from .. import core
from .load_toast_h5 import load_toast_h5_obs, load_toast_h5_dichroic_hack


# Interim, Oct. 2021.  Wait a year, remove this caution.
if hasattr(so3g, 'G3SuperTimestream'):
    from so3g import G3SuperTimestream
else:
    class MockG3SuperTimestream(object):
        pass
    G3SuperTimestream = MockG3SuperTimestream


class FieldGroup(list):
    """This is essentially a roadmap for decoding data from a
    G3FrameObject.  Each entry in this list is either a string, giving
    the key of a vector that should be loaded from the frame object,
    or a FieldGroup requesting a recursive decent into a "map" defined
    in the FrameObject.

    It is used to formulate decoding requests for a frame.  For
    example, a frame structured like this::

      Frame (Scan) [
        "boresight" (spt3g.core.G3TimestreamMap) => Timestreams from 3 detectors
        "signal" (spt3g.core.G3TimestreamMap) => Timestreams from 794 detectors
        "site_velocity" (spt3g.core.G3VectorDouble) => 3327 elements
      ]

    might have a roadmap defined like this::

      request = FieldGroup('root', [
        FieldGroup('boresight', ['az', 'el', 'roll']),
        FieldGroup('signal', ['det1', 'det2', 'det3']),
        'site_velocity'])

    Note that the name 'root', of the outermost FieldGroup, is not
    needed to decode a G3FrameObject... but you might use it for
    something else.

    When a string appears in the list, it is promoted to a "Field",
    internally.  If a field requires special processing, instantiate
    it directly to set options; e.g.::

      Field('site_velocity', optional=True)

    """
    def __init__(self, name, items=None, timestamp_field=None,
                 compression=False, refs_ok=True):
        """Arguments:
          name (str): The key in the parent G3FrameObject at which to
            find the specified items.
          items (list, optional): A list of items.
          timestamp_field (str): If not None, then the timestamps will
            be extracted from the indicated G3TimestreamMap and given
            the name specified here.
          compression (bool): If the requested item is a
            G3TimestreamMap with offset+gain compression implemented,
            then say so here.
          refs_ok (bool): If True, then extraction code will be
            instructed that it is ok to collect a bunch of 1-d
            references to the same large 2-d array, rather than
            copying out the 1-d subarrays of interest so big 2-d
            FrameObjects can be freed.  If False, copies will be
            forced (this is efficient in the limit that small number
            of available channels is being collected).

        """
        super().__init__()
        self.name = name
        self.compression = compression
        self.timestamp_field = timestamp_field
        self.refs_ok = refs_ok
        if items is not None:
            self.extend(items)

    def empty(self):
        """Returns an empty data structure (nested OrderederdDicts of lists)
        suitable for containing the data that this FieldGroup would
        decode.

        """
        output = OrderedDict()
        for item in self:
            if isinstance(item, FieldGroup):
                output[item.name] = item.empty()
                if item.timestamp_field:
                    output[item.timestamp_field] = []
            else:
                item = Field.as_field(item)
                output[item.name] = []
        return output

    @staticmethod
    def hstack_result(streams):
        """Recursively descends in streams (nested OrderedDicts) until it
        finds a list; assumes the list is a list of 1-d numpy arrays,
        and concatenates them together (in place).

        """
        keys = list(streams.keys())
        for k in keys:
            if isinstance(streams[k], list):
                streams[k] = np.hstack(streams[k])
            else:
                FieldGroup.hstack_result(streams[k])

    @staticmethod
    def merge_result(dest, src):
        """Recursively descends through src (nested OrderedDicts), and
        compares to dest.  In places where dest has the same keys, an
        update is performed on dest.  Any duplicate keys that are not
        OrderedDicts will be clobbered.

        For example, if the input is::

           dest = {'signal': {'a': [0,1,2,3],
                              'b': [1,2,3,4]},
                   'hk': {'temp1': [9,11,13,19]}}
           src =  {'signal': {'c': [2,3,4,5],
                              'd': [3,4,5,6]},
                   'hk': {'temp1': [9,11,13,19]}}

        Then dest will be updated to::

           dest = {'signal': {'a': [0,1,2,3],
                              'b': [1,2,3,4],
                              'c': [2,3,4,5],
                              'd': [3,4,5,6]},
                   'hk': {'temp1': [9,11,13,19]}}

        """
        keys = list(src.keys())
        for k in keys:
            if isinstance(src[k], dict):
                if not k in dest:
                    dest[k] = src[k]
                else:
                    FieldGroup.merge_result(dest[k], src[k])
            else:
                dest[k] = src[k]

class Field:
    def __init__(self, name, optional=False, wildcard=False, oversample=1):
        self.name = name
        self.wildcard = wildcard
        self.opts = {'optional': optional}
        self.oversample = oversample

    @staticmethod
    def as_field(item):
        if isinstance(item, Field):
            return item
        if isinstance(item, str):
            return Field(item)
        raise TypeError('Cannot promote %s to Field.' % item)


def unpack_frame_object(fo, field_request, streams, compression_info=None,
                        offset=0, max_count=None, refs_ok=True):
    """Unpack requested fields from a G3FrameObject, and update a data structure.
    
    Arguments:
      fo (G3FrameObject): The source object.
      field_request (FieldGroup): Description of what fields to unpack.
      streams (nested OrderedDicts): The data structure to unpack the data into.
      compression_info: The gain and offset dicts in the case of
        compression (or None to disable).

    """
    # Expand wildcards?
    to_remove = []
    to_add = []
    for item in field_request:
        if getattr(item, 'wildcard', False):
            assert(item.name == '*')  # That's the only wildcard we allow right now...
            to_remove.append(item)
            del streams[item.name]
            if isinstance(fo, G3SuperTimestream):
                for k in fo.names:
                    to_add.append(Field(k))
                    to_add[-1].opts = item.opts
                    streams[k] = []
            else:
                for k in fo.keys():
                    to_add.append(Field(k))
                    to_add[-1].opts = item.opts
                    streams[k] = []
    for item in to_remove:
        field_request.remove(item)
    field_request.extend(to_add)
    
    if isinstance(fo, G3SuperTimestream):
        key_map = {k: i for i, k in enumerate(fo.names)}
        
    def our_slice(n, oversamp):
        # returns (range size, slice)
        if n <= offset:
            return 0, slice(0, 0)
        n = n - offset
        if max_count is not None:
            n = min(n, max_count)
        return n, slice(offset * oversamp, (offset + n) * oversamp)

    _consumed = 0
    for item in field_request:
        if isinstance(item, FieldGroup):
            if item.compression:
                _gain = fo.get('compressor_gain_%s' % item.name, {})
                _offset = fo.get('compressor_offset_%s' % item.name, {})
                comp_info = (_gain, _offset)
            else:
                comp_info = None
            target = fo[item.name]
            _consumed = unpack_frame_object(target, item, streams[item.name], comp_info,
                                            offset=offset, max_count=max_count,
                                            refs_ok=item.refs_ok)
            _n, sl = our_slice(_consumed, 1)
            # Check and slice timestamp field independently -- must
            # work even if no dets requested led to _n=0 above.
            if item.timestamp_field is not None:
                if isinstance(target, G3SuperTimestream):
                    timesv = np.array(target.times) / g3core.G3Units.sec
                else:
                    t0, t1, ns = target.start, target.stop, target.n_samples
                    t0, t1 = t0.time / g3core.G3Units.sec, t1.time / g3core.G3Units.sec
                    timesv = np.linspace(t0, t1, ns)
                _consumed = len(timesv)
                _n, sl = our_slice(_consumed, 1)
                streams[item.timestamp_field].append(timesv[sl])
            continue
        # This is a simple field.
        item = Field.as_field(item)
        key = item.name
        if item.opts['optional']:
            if not key in fo.keys():
                if key in streams:
                    assert(len(streams[key]) == 0) # field went missing?
                    del streams[key]
                assert(key not in streams)
                continue
        if compression_info is not None:
            _gain, _offset = compression_info
            m, b = _gain.get(key, 1.), _offset.get(key, 0.)
            v = np.array(fo[key], dtype='float32') / m + b
        else:
            if isinstance(fo, G3SuperTimestream):
                v = fo.data[key_map[key]]
            else:
                v = np.array(fo[key])
        _consumed = len(v) // item.oversample
        _n, sl = our_slice(_consumed, item.oversample)

        if _n:
            if refs_ok:
                streams[key].append(v[sl])
            else:
                streams[key].append(np.copy(v[sl]))
    return _consumed

def unpack_frames(filename, field_request, streams, samples=None):
    """Read frames from the specified file and expand the data by stream.
    Only the requested fields, specified through *_fields arguments,
    are expanded.

    Arguments:
      filename (str): Full path to the file to load.
      field_request: Instructions for what fields to load.
      streams: Structure to which to append the
        streams from this file (perhaps obtained from running
        unpack_frames on a preceding file).
      samples (int, int): Start and end of sample range to unpack
        *from this file*.  First argument must be non-negative.  Second
        argument may be None, indicating to read forever.

    Returns:
      streams (structure containing lists of numpy arrays).

    """
    if streams is None:
        streams = field_request.empty()
    if samples is None:
        offset = 0
        to_read = None
    else:
        offset, to_read = samples
        if to_read is not None:
            to_read -= offset

    reader = so3g.G3IndexedReader(filename)
    while to_read is None or to_read > 0:
        frames = reader.Process(None)
        if len(frames) == 0:
            break
        frame = frames[0]
        if frame.type != g3core.G3FrameType.Scan:
            continue
        _consumed = unpack_frame_object(
            frame, field_request, streams, offset=offset, max_count=to_read)
        offset -= _consumed
        if offset < 0:
            if to_read is not None:
                to_read += offset
            offset = 0

    return streams


def load_file(filename, dets=None, signal_only=False):
    """Load data from file where there is no supporting obsfiledb.

    Args:
      filename (str or list): A filename or list of filenames (to be
        loaded in order).
      dets (list of str): The detector names of interest.  If None,
        loads all dets present in this file.  To load only the
        ancillary data, pass an empty list.
      signal_only (bool): If set, then only 'signal' is collected and
        other stuff is ignored.

    """
    if isinstance(filename, str):
        filenames = [filename]
    else:
        filenames = filename

    subreq = [
        FieldGroup('signal', [Field('*', wildcard=True)],
                   timestamp_field='timestamps', compression=True),
        ]
    subreq.extend([
        FieldGroup('boresight', ['az', 'el', 'roll']),
        Field('hwp_angle', optional=True),
        Field('corotator_angle', optional=True),
        Field('site_position', oversample=3),
        Field('site_velocity', oversample=3),
        Field('boresight_azel', oversample=4),
        Field('boresight_radec', oversample=4),
    ])
    request = FieldGroup('root', subreq)

    streams = None
    for filename in filenames:
        streams = unpack_frames(filename, request, streams)

    # Do we need to update that dets list?
    if dets is None:
        dets = list(streams['signal'].keys())

    # Create AxisManager now that we know the sample count.
    if len(dets) == 0:
        count = sum(map(len,streams['timestamps']))
    else:
        count = sum(map(len,streams['signal'][dets[0]]))
    aman = core.AxisManager(
        core.LabelAxis('dets', dets),
        core.OffsetAxis('samps', count, 0),
    )
    aman.wrap('signal', np.zeros(aman.shape, 'float32'),
              [(0, 'dets'), (1, 'samps')])

    if not signal_only:
        # The non-signal fields are duplicated across files so you
        # can just absorb them once.
        aman.wrap('timestamps', hstack_into(None, streams['timestamps']),
                      [(0, 'samps')])
        bman = core.AxisManager(aman.samps.copy())
        for k in ['az', 'el', 'roll']:
            bman.wrap(k, hstack_into(None, streams['boresight'][k]),
                      [(0, 'samps')])
        aman.wrap('boresight', bman)
        aman.wrap('qboresight_azel',
                  hstack_into(None, streams['boresight_azel']).reshape((-1, 4)),
                  [(0, 'samps')])
        aman.wrap('qboresight_radec',
                  hstack_into(None, streams['boresight_radec']).reshape((-1, 4)),
                  [(0, 'samps')])
        site = core.AxisManager(aman.samps.copy())
        for k in ['position', 'velocity']:
            site.wrap(k, hstack_into(None, streams['site_' + k]).reshape(-1, 3),
                      [(0, 'samps')])
        aman.wrap('site', site)

        if 'hwp_angle' in streams:
            aman.wrap('hwp_angle', hstack_into(None, streams['hwp_angle']),
                      [(0, 'samps')])

        if 'corotator_angle' in streams:
            aman.wrap('corotator_angle', hstack_into(None, streams['corotator_angle']),
                      [(0, 'samps')])

    # Copy in the signal, for each file.
    for det_name, arrs in streams['signal'].items():
        if dets is None or det_name in dets:
            i = list(aman.dets.vals).index(det_name)
            hstack_into(aman.signal[i], arrs)

    del streams
    return aman


def load_observation(db, obs_id, dets=None, samples=None, prefix=None,
                     no_signal=None,
                     **kwargs):
    """Obsloader function for TOAST simulate data -- this function matches
    output from pipe-s0001/s0002 (and SSO sims in 2019-2021).

    See API template, `sotodlib.core.context.obsloader_template`, for
    details.

    """
    if any([v is not None for v in kwargs.values()]):
        raise RuntimeError(
            f"This loader function does not understand kwargs: f{kwargs}")

    if no_signal is None:
        no_signal = False  # from here, assume no_signal in [True, False]

    # Regardless of what dets have been asked for (maybe none), get
    # the list of detsets implicated in this observation.
    c = db.conn.execute('select distinct DS.name, DS.det from detsets DS '
                        'join files on DS.name=files.detset '
                        'where obs_id=?', (obs_id,))
    pairs = [tuple(r) for r in c.fetchall()]

    # Now filter to only the dets requested.
    if dets is None:
        pairs_req = pairs
    else:
        pairs_req = [p for p in pairs if p[1] in dets]
        # Use sets for this...
        dets_req = [p[1] for p in pairs_req]
        unmatched = [d for d in dets if d not in dets_req]
        if len(unmatched):
            raise RuntimeError("User requested invalid dets (e.g. %s) "
                               "for obs_id=%s" % (unmatched[0], obs_id))

    # Group by detset.
    dets_by_detset = OrderedDict([(p[0], []) for p in pairs_req])
    for p in pairs_req:
        dets_by_detset[p[0]].append(p[1])

    # If user requested no dets, include one detset to get ancil.
    if len(dets_by_detset) == 0:
        dets_by_detset[pairs[0][0]] = []

    # Loop through relevant files, in sample order, and accumulate
    # lists of stream segments.
    aman = None

    for detset, dets in dets_by_detset.items():
        detset_files = db.get_files(obs_id, detsets=[detset], prefix=prefix)[detset]
        subreq = []
        if no_signal:
            if aman is None:
                subreq.extend([
                    FieldGroup('signal', [], timestamp_field='timestamps',
                               compression=True),
                ])
        else:
            subreq.extend([
                FieldGroup('signal', dets, timestamp_field='timestamps',
                           compression=True),
            ])
        if aman is None:
            subreq.extend([
                FieldGroup('boresight', ['az', 'el', 'roll']),
                Field('hwp_angle', optional=True),
                Field('corotator_angle', optional=True),
                Field('site_position', oversample=3),
                Field('site_velocity', oversample=3),
                Field('boresight_azel', oversample=4),
                Field('boresight_radec', oversample=4),
            ])
        if len(subreq) == 0:
            continue

        request = FieldGroup('root', subreq)

        if samples:
            sample_start, sample_stop = samples
        else:
            sample_start, sample_stop = 0, None

        stop = sample_stop

        file_list = db.get_files(obs_id, [detset], prefix=prefix)[detset]
        streams = None
        for row in detset_files:
            filename, file_start, file_stop = row
            assert(file_start is not None)
            start = max(0, sample_start - file_start)
            if sample_stop is not None:
                stop = sample_stop - file_start
            streams = unpack_frames(
                filename, request, streams, samples=(start, stop))

        if aman is None:
            # Create AxisManager now that we know the sample count.
            count = sum(map(len,streams['timestamps']))
            aman = core.AxisManager(
                core.LabelAxis('dets', [p[1] for p in pairs_req]),
                core.OffsetAxis('samps', count, sample_start, obs_id),
            )
            if no_signal:
                aman.wrap('signal', None)
            else:
                aman.wrap('signal', np.zeros(aman.shape, 'float32'),
                          [(0, 'dets'), (1, 'samps')])

            # The non-signal fields are duplicated across files so you
            # can just absorb them once.
            aman.wrap('timestamps', hstack_into(None, streams['timestamps']),
                          [(0, 'samps')])
            bman = core.AxisManager(aman.samps.copy())
            for k in ['az', 'el', 'roll']:
                bman.wrap(k, hstack_into(None, streams['boresight'][k]),
                          [(0, 'samps')])
            aman.wrap('boresight', bman)
            aman.wrap('qboresight_azel',
                      hstack_into(None, streams['boresight_azel']).reshape((-1, 4)),
                      [(0, 'samps')])
            aman.wrap('qboresight_radec',
                      hstack_into(None, streams['boresight_radec']).reshape((-1, 4)),
                      [(0, 'samps')])
            site = core.AxisManager(aman.samps.copy())
            for k in ['position', 'velocity']:
                site.wrap(k, hstack_into(None, streams['site_' + k]).reshape(-1, 3),
                          [(0, 'samps')])
            aman.wrap('site', site)

            if 'hwp_angle' in streams:
                aman.wrap('hwp_angle', hstack_into(None, streams['hwp_angle']),
                          [(0, 'samps')])

            if 'corotator_angle' in streams:
                aman.wrap('corotator_angle', hstack_into(None, streams['corotator_angle']),
                          [(0, 'samps')])

        # Copy in the signal, for each file.
        for det_name, arrs in streams['signal'].items():
            i = list(aman.dets.vals).index(det_name)
            hstack_into(aman.signal[i], arrs)

        del streams
    
    aman.wrap('flags', core.FlagManager.for_tod(aman))
    return aman


def hstack_into(dest, src_arrays):
    if dest is None:
        return np.hstack(src_arrays)
    offset = 0
    for ar in src_arrays:
        n = len(ar)
        dest[offset:offset+n] = ar
        offset += n
    return dest


# Register the loaders defined here.
core.OBSLOADER_REGISTRY.update(
    {
        'pipe-s0001': load_observation,
        'toast3-hdf': load_toast_h5_obs,
        'toast3-hdf-dichroic-hack': load_toast_h5_dichroic_hack,
        'default': load_observation,
    }
)


# Deprecated alias (used to live here...)
OBSLOADER_REGISTRY = core.OBSLOADER_REGISTRY



import os
import yaml
import time
import numpy as np
import argparse
import logging
from typing import Optional

from sotodlib import core
from sotodlib.io import hk_utils
from sotodlib.site_pipeline import util
logger = util.init_logger('make_cosamp_hk', 'make_cosamp_hk: ')

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()    
    parser.add_argument(
        'config',
        help="configuration yaml file.")
    parser.add_argument(
        '-o', '--output_dir', action='store', default=None, type=str,
        help='output data directory, overwrite config output_dir')
    parser.add_argument(
        '--verbose', default=2, type=int,
        help="increase output verbosity. \
        0: Error, 1: Warning, 2: Info(default), 3: Debug")
    parser.add_argument(
        '--overwrite',
        help="If true, overwrites existing entries in the database",
        action='store_true',
    )
    parser.add_argument(
        '--query_text', type=str,
        help="Query text",
    )
    parser.add_argument(
        '--query_tags', nargs="*", type=str,
        help="Query tags",
    )
    parser.add_argument(
        '--min_ctime', type=int,
        help="Minimum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--max_ctime', type=int,
        help="Maximum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--obs_id', type=str,
        help="obs_id of particular observation if we want to run on just one",
    )
    parser.add_argument(
        '--update-delay',
        help="Number of days (unit is days) in the past to start observation list.",
        type=int
    )
    return parser
    
def main(config: str,
        output_dir: Optional[str] = None,
        verbose: Optional[int] = 2,
        overwrite: Optional[bool] = False,
        query_text: Optional[str] = None,
        query_tags: Optional[list] = None,
        min_ctime: Optional[float] = None,
        max_ctime: Optional[float] = None,
        obs_id: Optional[str] = None,
        update_delay: Optional[str] = None,
        ):
    """
    Args:
        config: str
            Path to the configuration file containing processing parameters.
        output_dir: str or None
            Directory where the processed data will be saved. \
            If not provided, it is extracted from the configuration file.
        verbose: str
            Verbosity level for logging. Default is 2 (INFO).
        overwrite: bool
            Whether to overwrite existing output files. Default is False.
        query_text: str or None
            Text-based query to filter observations. If not provided, it is extracted from the configuration file.
        query_tags: list or None
            Tags used for further filtering observations. If not provided, it is extracted from the configuration file.
        min_ctime: float or None
            Minimum timestamp of observations to be queried. If not provided, it is extracted from the configuration file.
        max_ctime: float or None
            Maximum timestamp of observations to be queried. If not provided, it is extracted from the configuration file.
        obs_id: str or None
            Specific observation obs_id to process. If provided, overrides other filtering parameters.
        update_delay: str or None
            Number of days (unit is days) in the past to start observation list.
    """
    # Set verbose
    if verbose == 0:
        logger.setLevel(logging.ERROR)
    elif verbose == 1:
        logger.setLevel(logging.WARNING)
    elif verbose == 2:
        logger.setLevel(logging.INFO)
    elif verbose == 3:
        logger.setLevel(logging.DEBUG)
    
    # load configuration file
    if type(config) == str:
        with open(config) as f:
            config = yaml.safe_load(f)
    
    # load context
    context_file = config['context_file']
    ctx = core.Context(context_file)
    
    if output_dir is None:
        if 'output_dir' in config.keys():
            output_dir = config['output_dir']
        else:
            raise ValueError('output_dir is not specified.')
    if query_text is None:
        if 'query_text' in config.keys():
            query_text = config['query_text']
    if query_tags is None:
        if 'query_tags' in config.keys():
            query_tags = config['query_tags']
    if min_ctime is None:
        if 'min_ctime' in config.keys():
            min_ctime = config['min_ctime']
    if max_ctime is None:
        if 'max_ctime' in config.keys():
            max_ctime = config['max_ctime']
    if update_delay is None:
        if 'update_delay' in config.keys():
            update_delay = config['update_delay']
    
    full_fields = config['fields']
    aliases = config['aliases']
    data_fields = [full_field.split('.')[1] for full_field in full_fields]
            
    save_mean = config.get('save_mean', False)
    save_median = config.get('save_median', False)
    save_rms = config.get('save_rms', False)
    save_ptp = config.get('save_ptp', False)
    
    min_valid_value = config.get('min_valid_value', None)
    max_valid_value = config.get('max_valid_value', None)
    max_valid_dvalue_dt = config.get('max_valid_dvalue_dt', None)
    if np.all(np.array([min_valid_value is None,
                        max_valid_value is None,
                        max_valid_dvalue_dt is None])):
        save_nan_fraction = False
    else:
        save_nan_fraction = True

    if (min_ctime is None) and (update_delay is not None):
        min_ctime = int(time.time()) - update_delay*86400
    logger.info(f'min_ctime: {min_ctime}')
    
    # Load metadata sqlite
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_prefix = config['output_prefix']
    man_db_filename = os.path.join(output_dir, f'{output_prefix}.sqlite')
    if os.path.exists(man_db_filename):
        logger.info(f"Mapping {man_db_filename} for the \
                    archive index.")
        man_db = core.metadata.ManifestDb(man_db_filename)
    else:
        logger.info(f"Creating {man_db_filename} for the \
                     archive index.")
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('dataset')
           
        if save_mean:
            scheme.add_data_field(f'{output_prefix}_mean')
        if save_median:
            scheme.add_data_field(f'{output_prefix}_median')
        if save_rms:
            scheme.add_data_field(f'{output_prefix}_rms')
        if save_ptp:
            scheme.add_data_field(f'{output_prefix}_ptp')
        if save_nan_fraction:
            scheme.add_data_field(f'{output_prefix}_nan_fraction')
        
        man_db = core.metadata.ManifestDb(
            man_db_filename,
            scheme=scheme
        )
        
    
    # query observations
    if obs_id is not None:
        tot_query = f"obs_id=='{obs_id}'"
    else:
        tot_query = "and "
        if query_text is not None:
            tot_query += f"{query_text} and "
        if min_ctime is not None:
            tot_query += f"timestamp>={min_ctime} and "
        if max_ctime is not None:
            tot_query += f"timestamp<={max_ctime} and "
        tot_query = tot_query[4:-4]
        if tot_query == "":
            tot_query = "1"
    
    logger.info(f'tot_query: {tot_query}')
    obs_list= ctx.obsdb.query(tot_query, query_tags)
    
    # output cosampled data for each obs_id
    for obs in obs_list:
        obs_id = obs['obs_id']
        try:
            meta = ctx.get_meta(obs_id)
            aman = ctx.get_obs(obs_id, dets=[])
        except Exception as e:
            logger.error(f"Exception \n '{e}' \n thrown while aman loading of {obs_id}. Skipped")
            continue
            
        try:
            _hkman = hk_utils.get_detcosamp_hkaman(aman, 
                                      fields = config['fields'],
                                      data_dir = config['input_dir'])
            dt_buffer = 60
            start = float(aman.timestamps[0] - dt_buffer)
            stop = float(aman.timestamps[-1] + dt_buffer)
            _data = hk_utils.sort_hkdata(start=start, stop=stop, fields=config['fields'],
                                         data_dir=config['input_dir'], alias=None)
            _hkman = hk_utils.make_hkaman(grouped_data=_data, alias_exists=False,
                                          det_cosampled=True, det_aman=aman)
            hkman = core.AxisManager(meta.dets, aman.samps)
            hkman.wrap('timestamps', aman.timestamps, [(0, 'samps')])
            
            num_different_field = 0
            for i, (data_field, alias) in enumerate(zip(data_fields, aliases)):
                if i > 0:
                    num_different_field += int(data_fields[i-1] != data_field)
                i_in_same_field = i - num_different_field
                hkman.wrap(alias, _hkman[data_field][data_field][i_in_same_field], [(0, 'samps')])
            # set values nan if it is invalid
            mask_invalid = np.zeros(hkman.samps.count, dtype=bool)
            if min_valid_value is not None:
                mask_invalid[hkman[aliases[0]] < min_valid_value] = True
            if max_valid_value is not None:
                mask_invalid[max_valid_value < hkman[aliases[0]]] = True
            if max_valid_dvalue_dt is not None:
                dt = np.median(np.diff(aman.timestamps))
                mask_invalid[ np.abs( np.diff(np.append(hkman[aliases[0]], hkman[aliases[0]][-1])) / dt) > max_valid_dvalue_dt ] = True
                mask_invalid[ np.abs( np.diff(np.append(hkman[aliases[0]][0],  hkman[aliases[0]])) / dt) > max_valid_dvalue_dt ] = True
            if np.any(mask_invalid):
                hkman[aliases[0]][mask_invalid] = np.nan
                
        except Exception as e:
            logger.warning(f"Exception \n '{e}' \n thrown while hk loading of {obs_id}. Making dummy hkaman")
            hkman = core.AxisManager(meta.dets, aman.samps)
            hkman.wrap('timestamps', aman.timestamps, [(0, 'samps')])
            num_different_field = 0
            for i, (data_field, alias) in enumerate(zip(data_fields, aliases)):
                if i > 0:
                    num_different_field += int(data_fields[i-1] != data_field)
                i_in_same_field = i - num_different_field
                hkman.wrap(alias, np.nan*np.zeros(hkman.samps.count), [(0, 'samps')])
            
        try:
            h5_filename = f"{output_prefix}_{obs_id.split('_')[1][:4]}.h5"
            output_filename = os.path.join(output_dir, h5_filename)
            hkman.save(dest=output_filename, group=obs_id, overwrite=overwrite)
            
            entry_dict = {'obs:obs_id': obs_id, 'dataset': obs_id}
            if save_mean:
                entry_dict[f'{output_prefix}_mean'] = np.nanmean(hkman[aliases[0]])
            if save_median:
                entry_dict[f'{output_prefix}_median'] = np.nanmedian(hkman[aliases[0]])
            if save_rms:
                entry_dict[f'{output_prefix}_rms'] = np.nanstd(hkman[aliases[0]])
            if save_ptp:
                entry_dict[f'{output_prefix}_ptp'] = np.nanmax(hkman[aliases[0]]) - np.nanmin(hkman[aliases[0]])
            if save_nan_fraction:
                entry_dict[f'{output_prefix}_nan_fraction'] = np.mean(np.isnan(hkman[aliases[0]]).astype('float'))
            man_db.add_entry(entry_dict, filename=output_filename, replace=overwrite)
            logger.info(f"saved: {obs_id}")
        except Exception as e:
            logger.error(f"Exception \n '{e}' \n thrown while saving of {obs_id}. Skipped")
            continue
            
            


if __name__ == '__main__':
    util.main_launcher(main, get_parser)
    


from collections import OrderedDict as odict
import yaml
import os
import importlib
import logging
import numpy as np

from . import metadata
from .util import tag_substr
from .axisman import AxisManager, OffsetAxis, AxisInterface

logger = logging.getLogger(__name__)


class Context(odict):
    # Sets of special handlers may be registered in this class variable, then
    # requested by name in the context.yaml key "context_hooks".
    hook_sets = {}

    def __init__(self, filename=None, site_file=None, user_file=None,
                 data=None, load_list='all', metadata_list='all'):
        """Construct a Context object.  Note this is an ordereddict with a few
        attributes added on.

        Args:

          filename (str): Path to the dataset context file.  If None,
            that's fine.
          site_file (str): Path to the site file.  If None, then the
            value of SOTODLIB_SITECONFIG environment variable is used;
            unless that's unset in which case the file site.yaml in
            the current directory will be used.
          user_file (str): Path to the user file.  If None, then the
            value of SOTODLIB_USERCONFIG environment variable is used;
            unless that's unset in which case the file
            ~/.sotodlib.yaml will be used.
          data (dict or None): Optional dict of context data to merge
            in, after loading the site, user and main context files.
            Note the data are merged in with the usual rules (so items
            in data['tags'] will me merged into self['tags'].)
          load_list (str or list): A list of databases to load; some
            combination of 'obsdb', 'detdb', 'obsfiledb', or the
            string 'all' to load all of them (default).
          metadata_list (str or list): A list of metadata labels to load;
            some combinations of metadata labels, or the string 'all' to
            load all of them (default).

        """
        super().__init__()
        # Start with site and user config.
        site_ok, site_file, site_cfg = _read_cfg(
            site_file, 'SOTODLIB_SITECONFIG',
            os.path.join(os.getcwd(), 'site.yaml'))
        logger.info(f'Using site_file={site_file}.')
        user_ok, user_file, user_cfg = _read_cfg(
            user_file, 'SOTODLIB_USERCONFIG',
            os.path.expanduser('~/.sotodlib.yaml'))
        logger.info(f'Using user_file={user_file}.')

        self.update(site_cfg)
        self.update_context(user_cfg, metadata_list)

        ok, full_filename, context_cfg = _read_cfg(filename)
        if filename is not None and not ok:
            raise RuntimeError(
                'Could not load requested context file %s' % filename)
        logger.info(f'Using context_file={full_filename}.')
        self.update_context(context_cfg, metadata_list)

        # Update with anything the user passed in.
        if data is not None:
            self.update_context(data, metadata_list)

        self.site_file = site_file
        self.user_file = user_file
        self.filename = full_filename

        self.obsdb = None
        self.detdb = None
        self.obsfiledb = None
        self.obs_detdb = None

        for to_import in self.get('imports', []):
            importlib.import_module(to_import)

        # Activate the requested hook set
        if self.get('context_hooks'):
            self._hooks = self.hook_sets[self['context_hooks']]
        else:
            self._hooks = {}

        # Check-default 'tags' dict.
        self['tags'] = self._get_warn_missing('tags', {})

        # Perform recursive substitution on strings defined in tags.
        tag_substr(self, self['tags'])

        # Load basic databases.
        self.reload(load_list)

        # Call a post-processing hook before returning to user?
        self._call_hook('on-context-ready')

    def _call_hook(self, hook_key, *args, **kwargs):
        hook_func = self._hooks.get(hook_key)
        if hook_func is None:
            return
        logger.info('Calling hook for %s: %s' % (hook_key, hook_func))
        hook_func(self, *args, **kwargs)

    def _get_warn_missing(self, k, default=None):
        if not k in self:
            logger.warning(f'Key "{k}" not present in context.')
            return default
        return self[k]

    def update_context(self, new_stuff, metadata_list='all'):
        appendable = ['metadata']
        mergeable = ['tags']

        for k, v in new_stuff.items():
            if k == 'metadata':
                if metadata_list != 'all':
                    v = [m for m in v if 'label'
                         in m and m['label'] in metadata_list]
            if k in appendable and k in self:
                self[k].extend(v)
            elif k in mergeable and k in self:
                self[k].update(v)
            else:
                self[k] = v

    def reload(self, load_list='all'):
        """Load (or reload) certain databases associated with this dataset.
        (Note we don't load any per-observation metadata here.)

        """
        # Metadata support databases.
        for key, cls in [('detdb', metadata.DetDb),
                         ('obsdb', metadata.ObsDb),
                         ('obsfiledb', metadata.ObsFileDb)]:
            if (load_list == 'all' or key in load_list) and key in self:
                db_file = self[key]
                if not db_file.startswith('/'):
                    # Relative to context file.
                    db_file = os.path.join(os.path.split(self.filename)[0], db_file)
                db_file = os.path.abspath(db_file)
                logger.info(f'Loading {key} from {self[key]} -> {db_file}.')
                try:
                    db = cls.from_file(db_file, force_new_db=False)
                except Exception as e:
                    logger.error(f'DB failure when loading {key} from {self[key]} -> {db_file}\n')
                    raise e
                setattr(self, key, db)
        # Special overrides.
        if self.get('obsfiledb.prefix'):
            self.obsfiledb.prefix = self['obsfiledb.prefix']

        # The metadata loader.
        if load_list == 'all' or 'loader' in load_list:
            self.loader \
                = metadata.SuperLoader(self, obsdb=self.obsdb)

    def get_obs(self,
                obs_id=None,
                dets=None,
                samples=None,
                filename=None,
                detsets=None,
                meta=None,
                ignore_missing=None,
                on_missing=None,
                free_tags=None,
                no_signal=None,
                no_headers=None,
                special_channels=None,
                loader_type=None,
    ):
        """Load TOD and supporting metadata for some observation.

        Most arguments to this function are also accepted by (and in
        fact passed directly to) :func:`get_meta`, but are documented
        here.

        Args:
          obs_id (multiple): The observation to load (see Notes).
          dets (list, array, dict or ResultSet): The detectors to
            read.  If None, all dets will be read.
          samples (tuple of ints): The start and stop sample indices.
            If None, read all samples.  (Note that some loader
            functions might not support this argument.)
          filename (str): The path to a file to load, instead of using
            obs_id.  It is still required that this file appear in the
            obsfiledb, but this shortcut will automatically determine
            the obs_id and the detector and sample range selections
            that correspond to this single file.
          detsets (list, array): The detsets to read (with None
            equivalent to requesting all detsets).
          meta (AxisManager): An AxisManager returned by get_meta
            (though possibly with additional axis restrictions
            applied) to use as a starting point for detector selection
            and sample range.  (This will eventually be passed back to
            get_meta in the meta= argument, to fill in any missing
            metadata fields.)
          free_tags (list): Strings to match against the
            obs_colon_tags fields for detector restrictions.
          ignore_missing (bool): If True, don't fail when a metadata
            item can't be loaded, just try to proceed without it.
          on_missing (dict): If a metadata entry has a label that
            matches a key in this dict, the corresponding value in
            this dict will override the on_missing setting from the
            metadata entry.
          no_signal (bool): If True, the .signal will be set to None.
            This is a way to get the axes and pointing info without
            the (large) TOD blob.  Not all loaders may support this.
          no_headers (bool): If True, avoid loading "header"
            information that tags along with the signal.
          special_channels (bool): If True, load "special" readout
            channels that are normally skipped (e.g. fixed tones).
          loader_type (str): Name of the registered TOD loader
            function to use (this will override whatever is specified
            in context.yaml).

        Notes:
          It is acceptable to pass the ``obs_id`` argument by position
          (first), but all other arguments should be passed by
          keyword.

          The ``obs_id`` can be any of the following:

          - a string -- this is interpreted as the literal obs_id as
            used in the ObsDb and ObsFileDb.  Note however that this
            string may include "free tags" (see below).
          - a dict -- this is understood to be an ObsDb record, and
            the value under key 'obs_id' will be used as the obs_id
            (the other items will be ignored).
          - an AxisManager -- this is a short-hand for passing an
            object through meta=... .  I.e., ``get_obs(obs_is=axisman)``
            is treated the same way as ``get_obs(obs_id=None, meta=axisman)``.

          Detector subselection is achieved through the ``dets``
          argument.  If this is a dict, the keys must all be fields
          appearing in det_info.  Typically det_info will include at
          least readout_id and detset (this is the indexing
          information from ObsFileDb).  Some examples are::

            dets={'readout_id': ['det_00', 'det_01']}
            dets={'detset': 'wafer21'}
            dets={'band': ['f090']}
            dets={'detset': ['wafer21', 'wafer22'], 'band': ['f150']}

          Each value in ``dets`` can be a single item, or a list or
          numpy array of items.  The keys may include an optional
          'dets:' prefix.

          If ``dets`` is passed as a list or numpy array, that is
          equivalent to passing that value in through a dict with key
          'readout_id'; e.g.::

            dets=['det_00', 'det_01']

          You can instead pass a "det_info" ResultSet directly into
          the dets argument; that is equivalent to passing
          dets=det_info['readout_id'].  This is to accomodate the
          following sort of pattern::

            det_info = context.get_det_info(obs_id)
            det_info = det_info.subset(rows=(det_info['band'] == 'f090'))
            tod = context.get_obs(obs_id, dets=det_info)

          The sample range to load is determined by the samples
          argument.  Use Python start/stop indexing; for example
          samples=(0, -2) will try to read all but the last two
          samples and samples=(100, None) will read all samples except
          the first 100.

          When passing in ``meta``, the obs_id, detector list, and
          sample range will be extracted from that object.  It is an
          error to also specify ``obs_id``, ``dets``, ``samples``,
          ``filename``, or ``free_ags`` (but this could change).

        """
        meta = self.get_meta(obs_id=obs_id, dets=dets, samples=samples,
                             filename=filename, detsets=detsets, meta=meta,
                             free_tags=free_tags, ignore_missing=ignore_missing,
                             on_missing=on_missing)

        # Use the obs_id, dets, and samples from meta.
        obs_id = meta['obs_info']['obs_id']
        dets = list(meta.det_info['readout_id'])
        if samples is None and 'samps' in meta:
            samples = (meta.samps.offset, meta.samps.offset + meta.samps.count)

        # Make sure standard obsloaders are registered ...
        from ..io import load as _

        # Load TOD.
        if loader_type is None:
            loader_type = self.get('obs_loader_type', 'default')
        loader_func = OBSLOADER_REGISTRY[loader_type]  # Register your loader?
        aman = loader_func(self.obsfiledb, obs_id, dets=dets,
                           samples=samples, no_signal=no_signal,
                           no_headers=no_headers,
                           special_channels=special_channels)

        if aman is None:
            return meta
        if meta is not None:
            if 'det_info' in aman and 'det_info' in meta:
                # If the loader added det_info, then perform a special
                # merge.  Duplicate keys should be avoided, because
                # checking the values are the same is annoying.
                _det_info = aman['det_info']
                del aman['det_info']
                _det_info.restrict_axes([meta.dets])
                for k in meta.det_info._fields:
                    if k in _det_info._fields:
                        try:
                            check = np.all([meta['det_info'][k] ==_det_info[k]])
                            if check:
                                _det_info.move(k, None)
                                continue
                        except Exception as e:
                            pass
                        logger.error(f'Key "{k}" is present in det_info returned by '
                                     f'observation loader as well as in metadata '
                                     f'databases; The two versions are not '
                                     f'comparable. dropping the loader version.')
                        _det_info.move(k, None)
                meta.det_info.merge(_det_info)
            aman.merge(meta)
        return aman

    def get_meta(self,
                 obs_id=None,
                 dets=None,
                 samples=None,
                 filename=None,
                 detsets=None,
                 meta=None,
                 free_tags=None,
                 check=False,
                 ignore_missing=False,
                 on_missing=None,
                 det_info_scan=False
    ):
        """Load supporting metadata for an observation and return it in an
        AxisManager.

        The arguments shared with :func:`get_obs` (``obs_id``,
        ``dets``, ``samples``, ``filename``, ``detsets``, ``meta``,
        ``free_tags``, ``ignore_missing``, ``on_missing``) have the
        same meaning as in that function and are treated in the same
        way.

        Args:
          check (bool): If True, run in a check mode where an attempt
            is made to load each metadata entry, but the results are
            not kept and instead the function returns a report on what
            entries could / could not be loaded
          det_info_scan (bool): If True, only process the metadata
            entries that explicitly modify det_info.

        Returns:
          AxisManager with a .dets LabelAxis and .det_info and
          .obs_info entries.  If samples is specified, or if any
          metadata loads triggered its creation, then the .samps
          OffsetAxis is also created.

        Notes:
          When ``meta`` is passed in, it will be used to figure out
          the obs_id and detector and sample selections; however a new
          metadata AxisManager is returned.  Users should not rely on
          this; future improvements might modify meta in place, and
          try to re-use entries already present rather than loading
          them a second time.

        """
        def _warn_conflict(preamble, **kwargs):
            fails = {k: v for k, v in kwargs.items() if v is not None}
            if len(fails):
                logger.warning(f'{preamble}: arguments ignored: {fails}')

        free_tag_fields = self.get('obs_colon_tags', [])
        free_tags = list(free_tags) if free_tags else []

        if filename is not None:
            _warn_conflict(
                'Passing filename={filename} to get_obs with incompatible other args',
                obs_id=obs_id, detsets=detsets, samples=samples)
            # Resolve this to an obs_id / detset combo.
            info = self.obsfiledb.lookup_file(filename, resolve_paths=True)
            obs_id = info['obs_id']
            detsets = info['detsets']
            if info['sample_range'] is None or None in info['sample_range']:
                samples = None
                logger.warning('Due to incomplete ObsFileDb info, passing filename=... '
                               'will cause *all* files for the detset covered '
                               'by that file to be loaded.')
            else:
                samples = info['sample_range']

        # Handle some special cases for obs_id; at the end of this
        # checks and conversion, obs_id should be a string.

        if isinstance(obs_id, AxisManager):
            # Just move that to the meta argument.
            _warn_conflict(
                'Argument obs_id=<AxisManager> is incompatible with other args',
                meta=meta)
            obs_id, meta = None, obs_id

        elif isinstance(obs_id, dict):
            obs_id = obs_id['obs_id']  # You passed in a dict.

        elif isinstance(obs_id, str):
            # If the obs_id has colon-coded free tags, extract them.
            if ':' in obs_id:
                tokens = obs_id.split(':')
                obs_id = tokens[0]
                free_tags.extend(tokens[1:])

        if meta is not None:
            _warn_conflict(
                'Argument meta=<AxisManager> causes det/sample args to be ignored',
                samples=samples, dets=dets, detsets=detsets)
            obs_id = meta.obs_info['obs_id']
            dets = {'dets:readout_id': list(meta.dets.vals)}
            if 'samps' in meta:
                samples = meta.samps.offset, meta.samps.offset + meta.samps.count

        # Call a hook after preparing obs_id but before loading obs
        self._call_hook('before-use-detdb', obs_id=obs_id)

        # Identify whether we should use a detdb or an obs_detdb
        # If there is an obs_detdb, use that.
        # Otherwise, use whatever is in self.detdb, even if that is None.
        if self.obs_detdb is not None:
            detdb = self.obs_detdb
        else:
            detdb = self.detdb

        # Initialize det_info, starting with detdb.
        det_info = None
        if detdb is not None:
            det_info = detdb.props()

            # Backwards compatibility -- add "readout_id" if not found.
            if 'readout_id' not in det_info.keys:
                logger.warning('DetDb does not contain "readout_id"; aliasing from "name".')
                det_info.merge(metadata.ResultSet(
                    ['readout_id'], [(name,) for name in det_info['name']]))

        # Incorporate detset info from obsfiledb.
        detsets_info = self.obsfiledb.get_det_table(obs_id)
        det_info = metadata.merge_det_info(det_info, detsets_info)

        # Make the request for SuperLoader
        request = {'obs:obs_id': obs_id}
        if detsets is not None:
            request['dets:detset'] = detsets

        # Convert dets argument to request entry(s)
        if isinstance(dets, dict):
            for k, v in dets.items():
                if not k.startswith('dets:'):
                    k = 'dets:' + k
                if k in request:
                    raise ValueError(f'Duplicate specification of dets field "{k}"')
                request[k] = v
        elif isinstance(dets, metadata.ResultSet):
            request['dets:readout_id'] = dets['readout_id']
        elif hasattr(dets, '__getitem__'):
            # lists, tuples, arrays ...
            request['dets:readout_id'] = dets
        elif dets is not None:
            # Try a cast ...
            request['dets:readout_id'] = list(dets)

        metadata_list = self._get_warn_missing('metadata', [])
        meta = self.loader.load(metadata_list, request, det_info=det_info, check=check,
                                free_tags=free_tags, free_tag_fields=free_tag_fields,
                                det_info_scan=det_info_scan, ignore_missing=ignore_missing,
                                on_missing=on_missing)
        if check:
            return meta

        if samples is not None:
            if 'samps' in meta:
                meta.restrict('samps', slice(*samples))
            else:
                start, stop = samples
                assert(start >= 0 and stop >= 0)  # This could be loosened using obsfiledb
                axm = AxisManager(OffsetAxis('samps', stop - start, start, obs_id))
                meta = meta.merge(axm)
        return meta

    def get_det_info(self,
                     obs_id=None,
                     dets=None,
                     samples=None,
                     filename=None,
                     detsets=None,
                     meta=None,
                     free_tags=None,
                     on_missing=None):
        """Pass all arguments to :func:`get_meta(det_info_scan=True)`, and
        then return only the det_info, as a ResultSet.

        """
        if meta is None:
            meta = self.get_meta(obs_id=obs_id, dets=dets, samples=samples,
                                 filename=filename, detsets=detsets, free_tags=free_tags,
                                 on_missing=on_missing, det_info_scan=True)
        # Convert
        def _unpack(aman):
            items = []
            for k in aman.keys():
                if isinstance(aman[k], AxisManager):
                    sub_items = _unpack(aman[k])
                    for _k, _c in sub_items:
                        items.append((f'{k}.{_k}', _c))
                elif isinstance(aman[k], AxisInterface):
                    pass
                else:
                    items.append((k, aman[k]))
            return items

        items = _unpack(meta.det_info)
        return metadata.ResultSet([k for k, v in items],
                                  zip(*[v for k, v in items]))


def _read_cfg(filename=None, envvar=None, default=None):
    """Load a YAML file.  If filename is None, use the filename specified
    in the environment variable called envvar.  If that is not defined
    or decodes to None or an empty string, use the filename specified
    in default.

    Returns (ok, full_path, data) where ok is a boolean indicating
    whether the file at full_path was found on the file-system,
    full_path is the full path to the resolved filename (or None if
    not resolved), and data is the OrderedDict containing the data (or
    {} if not decoded).

    """
    if filename is None and envvar is not None:
        filename = os.getenv(envvar)
        if filename is None or filename == '':
            filename = None
    if filename is None and default is not None:
        filename = default
    if filename is None:
        return False, None, odict()
    filename = os.path.abspath(filename)
    if not os.path.exists(filename):
        return False, filename, odict()
    with open(filename, 'r') as file:
        yaml_file = yaml.safe_load(file)
    return True, filename, yaml_file


def obsloader_template(db, obs_id, dets=None, prefix=None, samples=None,
                       no_signal=None,
                       **kwargs):
    """This function is here to document the API for "obsloader" functions
    used by the Context system.  "obsloader" functions are used to
    load time-ordered detector data (rather than supporting metadata)
    from file archives, and return an AxisManager.

    Args:
      db (ObsFileDB): The database supporting the data files.
      obs_id (str): The obs_id (as recognized by ObsFileDb).
      dets (list of str): The dets to load.  If None, all dets are
        loaded.  If an empty list, ancillary data for the observation
        is still loaded.
      samples (tuple): The (start, end) indices of samples which
        should be loaded.  If start is None, 0 is used.  If end is
        None, sample_count is used.  Passing None is equivalent to
        passing (None, None).
      prefix (str): The root address of the data files, if not already
        known to the ObsFileDb.  (This is passed through to ObsFileDb
        prefix= argument.)
      no_signal (bool): If True, loader should avoid reading signal
        data (if possible) and should set .signal=None in the output.
        Passing None is equivalent to passing False.
      no_headers (bool): If True, loader should avoid loading "header"
        information that tags along with the signal (such as frame
        counters, biases, etc.)
      special_channels (bool): If True, load additional special /
        diagnostic readout channels that are normally skipped.

    Notes:
      This interface is subject to further extension.  When possible
      such extensions should take the form of optional arguments,
      whose default value is None and which are not activated except
      when needed.  This permits existing loaders to future-proof
      themselves by including ``**kwargs`` in the function signature
      but raising an exception if kwargs contains anything strange.
      See the body of this example function for template code to
      reject unexpected kwargs.

    Returns:
      An AxisManager with the data.

    """
    if any([v is not None for v in kwargs.values()]):
        raise RuntimeError(
            f"This loader function does not understand these kwargs: f{kwargs}")
    raise NotImplementedError("This is just a template function.")


#: OBSLOADER_REGISTRY will be accessed by the Context system to load
#: TOD.  The function obsloader_template, in this module, shows the
#: signature and describes the interface.

OBSLOADER_REGISTRY = {}



import numpy as np
import so3g

class PmatCut:
    """Implementation of cuts-as-extra-degrees-of-freedom for a single obs."""
    def __init__(self, cuts, model=None, params={"resolution":100, "nmax":100}):
        self.cuts   = cuts
        self.model  = model or "full"
        self.params = params
        self.njunk  = so3g.process_cuts(self.cuts.ranges, "measure", self.model, self.params, None, None)

    def forward(self, tod, junk):
        """Project from the cut parameter (junk) space for this scan to tod."""
        so3g.process_cuts(self.cuts.ranges, "insert", self.model, self.params, tod, junk)

    def backward(self, tod, junk):
        """Project from tod to cut parameters (junk) for this scan."""
        so3g.process_cuts(self.cuts.ranges, "extract", self.model, self.params, tod, junk)
        self.clear(tod)

    def clear(self, tod):
        junk = np.empty(self.njunk, tod.dtype)
        so3g.process_cuts(self.cuts.ranges, "clear", self.model, self.params, tod, junk)